import sys
import warnings

import numpy as np
from scipy.stats import rankdata


TAB = '    '
maxfloat = np.float128 if hasattr(np, 'float128') else np.longdouble
_slsqp_epsilon = np.sqrt(np.finfo(float).eps)  # scipy's default value for the SLSQP epsilon parameter


class ReprMixin:
    def __repr__(self):
        return f'{self.__class__.__name__}\n' + '\n'.join([f'\t{k}: {v}' for k, v in self.__dict__.items()])


def _check_param(x):
    if hasattr(x, '__len__'):
        if len(x) == 2:
            return x
        elif len(x) == 1:
            return [x[0], x[0]]
        else:
            print(f'Something went wrong, parameter array has length {len(x)}')
    else:
        return [x, x]


def _check_criteria(x):
    if hasattr(x[0], '__len__'):
        return x
    else:
        return [x, x]


def pearson2d(x, y):
    x, y = np.asarray(x), np.asarray(y)
    mx, my = np.nanmean(x, axis=-1), np.nanmean(y, axis=-1)
    xm, ym = x - mx[..., None], y - my[..., None]
    r_num = np.nansum(xm * ym, axis=-1)
    r_den = np.sqrt(np.nansum(xm ** 2, axis=-1) * np.nansum(ym ** 2, axis=-1))
    r = r_num / r_den
    return r


def spearman2d(x, y, axis=0):
    x, y = np.asarray(x), np.asarray(y)
    xr, yr = rankdata(x, axis=axis), rankdata(y, axis=axis)
    mxr, myr = np.nanmean(xr, axis=-1), np.nanmean(yr, axis=-1)
    xmr, ymr = xr - mxr[..., None], yr - myr[..., None]
    r_num = np.nansum(xmr * ymr, axis=-1)
    r_den = np.sqrt(np.nansum(xmr ** 2, axis=-1) * np.nansum(ymr ** 2, axis=-1))
    r = r_num / r_den
    return r


def print_warnings(w):
    for el in set([w_.message.args[0] for w_ in w]):
        if 'delta_grad == 0.0' not in el:
            print('\tWarning: ' + el)


def raise_warning_in_catch_block(msg, category, w):
    warnings.warn(msg, category=category)
    if len(w):
        sys.stderr.write(warnings.formatwarning(
            w[-1].message, w[-1].category, w[-1].filename, w[-1].lineno, line=w[-1].line
        ))


def type2roc(correct, conf, nbins=5):
    # Calculate area under type 2 ROC
    #
    # correct - vector of 1 x ntrials, 0 for error, 1 for correct
    # conf - vector of continuous confidence ratings between 0 and 1
    # nbins - how many bins to use for discretization

    bs = 1 / nbins
    h2, fa2 = np.full(nbins, np.nan), np.full(nbins, np.nan)
    for c in range(nbins):
        if c:
            h2[nbins - c - 1] = np.sum((conf > c*bs) & (conf <= (c+1)*bs) & correct.astype(bool)) + 0.5
            fa2[nbins - c - 1] = np.sum((conf > c*bs) & (conf <= (c+1)*bs) & ~correct.astype(bool)) + 0.5
        else:
            h2[nbins - c - 1] = np.sum((conf >= c * bs) & (conf <= (c + 1) * bs) & correct.astype(bool)) + 0.5
            fa2[nbins - c - 1] = np.sum((conf >= c * bs) & (conf <= (c + 1) * bs) & ~correct.astype(bool)) + 0.5

    h2 /= np.sum(h2)
    fa2 /= np.sum(fa2)
    cum_h2 = np.hstack((0, np.cumsum(h2)))
    cum_fa2 = np.hstack((0, np.cumsum(fa2)))

    k = np.full(nbins, np.nan)
    for c in range(nbins):
        k[c] = (cum_h2[c+1] - cum_fa2[c])**2 - (cum_h2[c] - cum_fa2[c+1])**2

    auroc2 = 0.5 + 0.25*np.sum(k)

    return auroc2



import os
import pathlib
import pickle
import timeit
import warnings
from dataclasses import make_dataclass

import numpy as np
from scipy.optimize import minimize, OptimizeResult
from scipy.stats import logistic as logistic_dist, binom

from .configuration import Configuration
from .dist import get_dist
from .fit import fmincon
from .gendata import simu_data
from .modelspec import Model, Data
from .plot import plot_link_function, plot_confidence_dist
from .transform import warp, noise_meta_transform, noise_sens_transform, logistic, link_function, link_function_inv
from .util import _check_param, TAB
from .util import maxfloat

np.set_printoptions(suppress=True)


class ReMeta:

    def __init__(self, cfg=None, **kwargs):
        """
        Main class of the ReMeta toolbox

        Parameters
        ----------
        cfg : util.Configuration
            Configuration object. If None is passed, the default configuration is used (but see kwargs).
        kwargs : dict
            The kwargs dictionary is parsed for keywords that match keywords of util.Configuration; in case of a match,
            the configuration is set.
        """

        if cfg is None:
            # Set configuration attributes that match keyword arguments
            cfg_kwargs = {k: v for k, v in kwargs.items() if k in Configuration.__dict__}
            self.cfg = Configuration(**cfg_kwargs)
        else:
            self.cfg = cfg
        self.cfg.setup()

        if self.cfg.meta_noise_dist.startswith('truncated_') and self.cfg.meta_noise_dist.endswith('_lookup'):
            try:
                self.lookup_table = np.load(f"lookup_{self.cfg.meta_noise_dist}_{self.cfg.meta_noise_type}.npz")
            except FileNotFoundError:
                raise FileNotFoundError('Lookup table not found. Lookup tables are not deployed via pip. You can '
                                        'download them from Github and put in a directory named "lookup"')
        else:
            self.lookup_table = None

        self.model = Model(cfg=self.cfg)
        self.data = None

        # self._punish_message = False

        self.sens_is_fitted = False
        self.meta_is_fitted = False

        # Define the function that is minimized for metacognitive parameter fitting
        self.fun_meta = dict(noisy_report=self._negll_meta_noisyreport,
                             noisy_readout=self._negll_meta_noisyreadout)[self.cfg.meta_noise_type]
        self.fun_meta_helper = dict(noisy_report=self._helper_negll_meta_noisyreport,
                                    noisy_readout=self._helper_negll_meta_noisyreadout)[self.cfg.meta_noise_type]

    def fit(self, stimuli, choices, confidence, precomputed_parameters=None, guess_meta=None, verbose=True,
            ignore_warnings=False):
        """
        Fit sensory and metacognitive parameters

        Parameters
        ----------
        stimuli : array-like of shape (n_samples)
            Array of signed stimulus intensity values, where the sign codes the stimulus category (cat 1: -, cat2: +)
            and the absolut value codes the intensity. Must be normalized to [-1; 1], or set
            `normalize_stimuli_by_max=True`.
        choices : array-like of shape (n_samples)
            Array of choices coded as 0 (or alternatively -1) for the negative stimuli category and 1 for the positive
            stimulus category.
        confidence : array-like of shape (n_samples)
            Confidence ratings; must be normalized to the range [0;1].
        precomputed_parameters : dict
            Provide pre-computed parameters. A dictionary with all parameters defined by the model must be passed. This
            can sometimes be useful to obtain information from the model without having to fit the model.
            [ToDO: which information?]
        guess_meta : array-like of shape (n_params_meta)
            For testing: provide an initial guess for the optimization of the metacognitive level
        verbose : bool
            If True, information of the model fitting procedure is printed.
        ignore_warnings : bool
            If True, warnings during model fitting are supressed.
        """

        # Instantiate util.Data object and perform preprocessing of the data
        self.data = Data(self.cfg, stimuli, choices, confidence)
        self.data.preproc()

        if verbose:
            print('\n+++ Sensory level +++')
        # with warnings.catch_warnings(record=True) as w:
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=UserWarning, module='scipy.optimize',
                                    message='delta_grad == 0.0. Check if the approximated function is linear. If the '
                                            'function is linear better results can be obtained by defining the Hessian '
                                            'as zero instead of using quasi-Newton approximations.')
            if ignore_warnings:
                warnings.filterwarnings('ignore')
            if isinstance(precomputed_parameters, dict):
                if not np.all([p in precomputed_parameters for p in self.cfg.paramset_sens.names]):
                    raise ValueError('Set of precomputed sensory parameters is incomplete.')
                self.model.fit.fit_sens = OptimizeResult(
                    x=[precomputed_parameters[p] for p in self.cfg.paramset_sens.names],
                    fun=self._negll_sens([precomputed_parameters[p] for p in self.cfg.paramset_sens.names])
                )
            else:
                if self.cfg.paramset_sens.nparams > 0:
                    self.cfg.paramset_sens.constraints = self.cfg.constraints_sens_callable(self)
                    if verbose:
                        negll_initial_guess = self._negll_sens(self.cfg.paramset_sens.guess)
                        print(f'Initial guess (neg. LL: {negll_initial_guess:.2f})')
                        for i, p in enumerate(self.cfg.paramset_sens.names):
                            print(f'{TAB}[guess] {p}: {self.cfg.paramset_sens.guess[i]:.4g}')
                        print('Performing local optimization')
                    t0 = timeit.default_timer()
                    self.model.fit.fit_sens = minimize(
                        self._negll_sens, self.cfg.paramset_sens.guess, bounds=self.cfg.paramset_sens.bounds,
                        constraints=self.cfg.paramset_sens.constraints, method='trust-constr'
                    )
                    if self.cfg.enable_thresh_sens:
                        fit_powell = minimize(
                            self._negll_sens, self.cfg.paramset_sens.guess, bounds=self.cfg.paramset_sens.bounds,
                            constraints=self.cfg.paramset_sens.constraints, method='Powell'
                        )
                        if fit_powell.fun < self.model.fit.fit_sens.fun:
                            self.model.fit.fit_sens = fit_powell
                    self.model.fit.fit_sens.execution_time = timeit.default_timer() - t0

                else:
                    self.model.fit.fit_sens = OptimizeResult(x=None)
            if isinstance(self.cfg.true_params, dict):
                if not np.all([p in self.cfg.true_params for p in self.cfg.paramset_sens.base_names]):
                    raise ValueError('Set of provided true sensory parameters is incomplete.')
                psens_true = sum([[self.cfg.true_params[p]] if n == 1 else self.cfg.true_params[p] for n, p in
                                  zip(self.cfg.paramset_sens.base_len, self.cfg.paramset_sens.base_names)], [])
                self.model.fit.fit_sens.negll_true = self._negll_sens(psens_true)

            # call once again with final=True to save the model fit
            negll, params_sens, choiceprob, posterior, stimuli_final = \
                self._negll_sens(self.model.fit.fit_sens.x, final=True)
            if 'thresh_sens' in params_sens and params_sens['thresh_sens'] < self.data.stimuli_min:
                warnings.warn('Fitted threshold is below the minimal stimulus intensity; consider disabling '
                              'the sensory threshold by setting enable_thresh_sens to 0', category=UserWarning)
            self.model.store_sens(negll, params_sens, choiceprob, posterior, stimuli_final,
                                  stimulus_norm_coefficent=self.data.stimuli_unnorm_max)
            self.model.report_fit_sens(verbose)
            self.sens_is_fitted = True

        # if not ignore_warnings and verbose:
        #     print_warnings(w)

        if not self.cfg.skip_meta:

            # compute decision values
            self._compute_dv_sens()

            if verbose:
                print('\n+++ Metacognitive level +++')

            args_meta = [ignore_warnings, None]
            if precomputed_parameters is not None:
                if not np.all([p in precomputed_parameters for p in self.cfg.paramset_meta.names]):
                    raise ValueError('Set of precomputed metacognitive parameters is incomplete.')
                self.model.params_meta = {p: precomputed_parameters[p] for p in self.cfg.paramset_meta.names}
                self.model.fit.fit_meta = OptimizeResult(
                    x=[precomputed_parameters[p] for p in self.cfg.paramset_meta.names],
                    fun=self.fun_meta([precomputed_parameters[p] for p in self.cfg.paramset_meta.names])
                )
                fitinfo_meta = self.fun_meta(list(self.model.params_meta.values()), *args_meta, final=True)  # noqa
                self.model.store_meta(*fitinfo_meta)
            else:
                with warnings.catch_warnings(record=True) as w:  # noqa
                    warnings.filterwarnings('ignore', module='scipy.optimize')
                    # prepare constraints for metacognitive parameters (which may depend on decision values)
                    if self.cfg.paramset_meta.nparams > 0:
                        self.cfg.paramset_meta.constraints = self.cfg.constraints_meta_callable(self)
                        self.model.fit.fit_meta = fmincon(
                            self.fun_meta, self.cfg.paramset_meta, args_meta, gradient_free=self.cfg.gradient_free,
                            gridsearch=self.cfg.gridsearch, grid_multiproc=self.cfg.grid_multiproc,
                            global_minimization=self.cfg.global_minimization,
                            fine_gridsearch=self.cfg.fine_gridsearch,
                            gradient_method=self.cfg.gradient_method, slsqp_epsilon=self.cfg.slsqp_epsilon,
                            init_nelder_mead=self.cfg.init_nelder_mead,
                            guess=guess_meta,
                            verbose=verbose
                        )
                    else:
                        self.model.fit.fit_meta = OptimizeResult(x=None)

                # call once again with final=True to save the model fit
                fitinfo_meta = self.fun_meta(self.model.fit.fit_meta.x, *args_meta, final=True)
                self.model.store_meta(*fitinfo_meta)

            if self.cfg.true_params is not None:
                params_true_meta = sum([[self.cfg.true_params[p]] if n == 1 else self.cfg.true_params[p] for n, p in
                                        zip(self.cfg.paramset_meta.base_len, self.cfg.paramset_meta.base_names)],
                                       [])
                self.model.fit.fit_meta.negll_true = self.fun_meta(params_true_meta)

            self.model.report_fit_meta(verbose)

            self.model.params = {**self.model.params_sens, **self.model.params_meta}

            self.meta_is_fitted = True

            # if not ignore_warnings:
            #     print_warnings(w)

    def summary(self, extended=False, generative=True, generative_nsamples=1000):
        """
        Provides information about the model fit.

        Parameters
        ----------
        extended : bool
            If True, store various model variables in the summary object.
        generative : bool
            If True, compare model predictions of confidence with empirical confidence by repeatedly sampling from
            the generative model.
        generative_nsamples : int
            Number of samples used for the generative model (higher = more accurate).

        Returns
        ----------
        summary : dataclass
            Information about model fit.
        """

        if not self.cfg.skip_meta and generative:
            gen = simu_data(generative_nsamples, self.data.nsamples, self.model.params,
                            cfg=self.cfg, stimuli_ext=self.data.stimuli, verbose=False)
        else:
            gen = None
        summary_model = self.model.summary(extended=extended, fun_meta=self.fun_meta,
                                           confidence_gen=None if gen is None else gen.confidence,  # noqa
                                           confidence_emp=self.data.confidence if generative else None)
        desc = dict(data=self.data.summary(extended), model=summary_model, cfg=self.cfg)

        summary_ = make_dataclass('Summary', desc.keys())

        def repr_(self_):
            txt = f'***{self_.__class__.__name__}***\n'
            for k, v in self_.__dict__.items():
                if k == 'cfg':
                    txt += f"\n{k}: {type(desc['cfg'])} <not displayed>"
                else:
                    txt += f"\n{k}: {v}"
            return txt

        summary_.__repr__ = repr_
        summary_.__module__ = '__main__'
        summary = summary_(**desc)
        return summary

    def _negll_sens(self, params, final=False, return_noise=False):
        """
        Minimization function for the sensory level

        Parameters:
        -----------
        params : array-like of shape (nparams)
            Parameter array of the sensory level.
        final : bool
            If True, store latent variables and parameters.
        return_noise : bool
            If True, only return the sensory noise array (auxiliarly option to define constraints on the noise array).

        Returns:
        --------
        negll: float
            Negative (summed) log likelihood.
        """

        bl = self.cfg.paramset_sens.base_len
        params_sens = {p: params[int(np.sum(bl[:i]))] if n == 1 else [params[int(np.sum(bl[:i])) + j] for j in range(n)]
                       for i, (p, n) in enumerate(zip(self.cfg.paramset_sens.base_names, bl))}
        thresh_sens = _check_param(params_sens['thresh_sens'] if self.cfg.enable_thresh_sens else 0)
        bias_sens = _check_param(params_sens['bias_sens'] if self.cfg.enable_bias_sens else 0)

        if self.cfg.enable_warping_sens:
            stimuli_final = warp(self.data.stimuli, params_sens['warping_sens'], self.cfg.function_warping_sens)
        else:
            stimuli_final = self.data.stimuli

        cond_neg, cond_pos = stimuli_final < 0, stimuli_final >= 0
        dv_sens = np.full(stimuli_final.shape, np.nan)
        # dv_sens[cond_neg] = (np.abs(stimuli_final[cond_neg]) > thresh_sens[0]) * \
        #                   (stimuli_final[cond_neg] - np.sign(stimuli_final[cond_neg]) * thresh_sens[0]) - bias_sens[0]
        # dv_sens[cond_pos] = (np.abs(stimuli_final[cond_pos]) > thresh_sens[1]) * \
        #                   (stimuli_final[cond_pos] - np.sign(stimuli_final[cond_pos]) * thresh_sens[1]) - bias_sens[1]
        dv_sens[cond_neg] = (np.abs(stimuli_final[cond_neg]) > thresh_sens[0]) * stimuli_final[cond_neg] + bias_sens[0]
        dv_sens[cond_pos] = (np.abs(stimuli_final[cond_pos]) > thresh_sens[1]) * stimuli_final[cond_pos] + bias_sens[1]

        noise_sens = self._noise_sens_transform(stimuli_final, params_sens)
        if return_noise:
            return noise_sens

        if self.cfg.detection_model:
            p_active = np.tanh(np.abs(dv_sens) / noise_sens)  # probability that a channel is active
            dv_congruent = np.sign(dv_sens) == np.sign(stimuli_final)
            p_correct = (dv_congruent * (1 - 0.5 * (1 - p_active) ** self.cfg.detection_model_nchannels) +
                         ~dv_congruent * 0.5 * (1 - p_active) ** self.cfg.detection_model_nchannels)
            posterior = (self.data.stimulus_ids == 1) * p_correct + (self.data.stimulus_ids == 0) * (1 - p_correct)
        else:
            posterior = logistic(dv_sens, noise_sens)
        choiceprob = (self.data.choices == 1) * posterior + (self.data.choices == 0) * (1 - posterior)
        negll = np.sum(-np.log(np.maximum(choiceprob, self.cfg.min_likelihood_sens)))

        return (negll, params_sens, choiceprob, posterior, stimuli_final) if final else negll

    def _negll_meta_noisyreadout(self, params, ignore_warnings=False, mock_binsize=None, final=False,
                                 return_noise=False, return_criteria=False, return_levels=False,
                                 constraint_mode=False):
        """
        Minimization function for the noisy-report model

        Parameters:
        -----------
        params : array-like of shape (nparams)
            Parameter array of the metacognitive level.
        ignore_warnings : bool
            If True, suppress warnings during minimization.
        mock_binsize : float
            Binsize around empirical confidence ratings to evaluate the likelihood. If not None, also returns the
            likelihood variable.
        final : bool
            If True, return latent variables and parameters.
        return_noise : bool
            If True, only return the metacognitive noise array (auxiliarly option to define constraints on the noise
            array).
        return_criteria : bool
            If True, only return confidence criteria (auxiliarly option to define constraints on confidence criteria).
        return_levels : bool
            If True, only return confidence levels (auxiliarly option to define constraints on confidence levels).
        constraint_mode : bool
            If True, method runs during scipy optimize constraint testing.

        Returns:
        --------
        By default, the method returns the negative log likelihood. However, depending on the arguments various
        combinations of variables are returned (see Parameters).
        negll: float
            Negative (summed) log likelihood.
        """

        params_meta, noise_meta, dist, dv_meta_considered, likelihood, likelihood_pdf, criteria_meta, levels_meta, \
            punishment_factor = \
            self._helper_negll_meta_noisyreadout(params, ignore_warnings, mock_binsize, final, return_noise,
                                                 return_criteria, return_levels, constraint_mode)

        # compute log likelihood
        likelihood_weighted_cum = np.nansum(self.model.dv_sens_pmf * likelihood, axis=1)
        if self.cfg.experimental_likelihood:
            # use an upper bound for the negative log likelihood based on a uniform 'guessing' model
            negll = min(self.model.max_negll, -np.sum(np.log(np.maximum(likelihood_weighted_cum, 1e-200))))
        else:
            negll = -np.sum(np.log(np.maximum(likelihood_weighted_cum, self.cfg.min_likelihood_meta)))
        negll *= punishment_factor

        if final:
            self.model.confidence = self._link_function(dv_meta_considered, params_meta,
                                                        criteria_meta=criteria_meta, levels_meta=levels_meta)
            return negll, params_meta, noise_meta, likelihood, dv_meta_considered, likelihood_weighted_cum, \
                   likelihood_pdf  # noqa
        elif mock_binsize is not None:
            return negll, likelihood
        else:
            return negll

    def _helper_negll_meta_noisyreadout(self, params, ignore_warnings=False, mock_binsize=None, final=False,
                                        return_noise=False, return_criteria=False, return_levels=False,
                                        constraint_mode=False):  # noqa
        """
        Minimization function for the noisy-readout model

        Parameters:
        -----------
        params : array-like of shape (nparams)
            Parameter array of the metacognitive level.
        ignore_warnings : bool
            If True, suppress warnings during minimization.
        mock_binsize : float
            Binsize around empirical confidence ratings to evaluate the likelihood. If not None, also returns the
            likelihood variable.
        final : bool
            If True, return latent variables and parameters.
        return_noise : bool
            If True, only return the metacognitive noise array (auxiliarly option to define constraints on the noise
            array).
        return_criteria : bool
            If True, only return confidence criteria (auxiliarly option to define constraints on confidence criteria).
        return_levels : bool
            If True, only return confidence levels (auxiliarly option to define constraints on confidence levels).
        constraint_mode : bool
            If True, method runs during scipy optimize constraint testing.

        Returns:
        --------
        By default, the method returns the negative log likelihood. However, depending on the arguments various
        combinations of variables are returned (see Parameters).
        negll: float
            Negative (summed) log likelihood.
        """

        bl = self.cfg.paramset_meta.base_len
        params_meta = {p: params[int(np.sum(bl[:i]))] if n == 1 else [params[int(np.sum(bl[:i])) + j] for j in range(n)]
                       for i, (p, n) in enumerate(zip(self.cfg.paramset_meta.base_names, bl))}

        punishment_factor = 1
        if '_criteria' in self.cfg.meta_link_function:
            criteria_meta, levels_meta = self._prepare_criteria(params_meta)
            if return_criteria:
                return criteria_meta
            if return_levels:
                return levels_meta
            if not final:
                bds = [(params[i] >= b[0]) & (params[i] <= b[1]) for i, b in enumerate(self.cfg.paramset_meta.bounds)]
                cns = [con['fun'](params) >= 0 for con in self.cfg.paramset_meta.constraints]
                if not np.all(cns) or not np.all(bds):
                    # if not self._punish_message:
                    #     print('punish!')
                    #     self._punish_message = True
                    punishment_factor = 10
        else:
            criteria_meta, levels_meta = None, None

        dv_meta_considered = self._compute_dv_meta(params_meta)

        noise_meta = self._noise_meta_transform(dv_meta_considered, params_meta, ignore_warnings=ignore_warnings)
        if return_noise:
            return noise_meta

        binsize = self.cfg.binsize_meta if mock_binsize is None else mock_binsize
        if self.cfg.experimental_wrap_binsize_meta:
            wrap_neg = binsize - np.abs(np.minimum(1, self.data.confidence_2d + binsize) - self.data.confidence_2d)  # noqa
            wrap_pos = binsize - np.abs(np.maximum(0, self.data.confidence_2d - binsize) - self.data.confidence_2d)  # noqa
            binsize_neg, binsize_pos = binsize + wrap_neg, binsize + wrap_pos
        else:
            binsize_neg, binsize_pos = binsize, binsize
        dv_meta_from_conf_lb = self._link_function_inv(
            np.maximum(0, self.data.confidence_2d - binsize_neg), self.cfg.meta_link_function, params_meta,
            criteria_meta=criteria_meta, levels_meta=levels_meta)
        dv_meta_from_conf_ub = self._link_function_inv(
            np.minimum(1, self.data.confidence_2d + binsize_pos), self.cfg.meta_link_function, params_meta,
            criteria_meta=criteria_meta, levels_meta=levels_meta)
        dv_meta_from_conf = self._link_function_inv(
            self.data.confidence_2d, self.cfg.meta_link_function, params_meta,
            criteria_meta=criteria_meta, levels_meta=levels_meta)

        dist = get_dist(self.cfg.meta_noise_dist, mode=dv_meta_considered, scale=noise_meta,
                        meta_noise_type='noisy_readout', lookup_table=self.lookup_table)
        if self.cfg.meta_noise_dist.startswith('censored_'):
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', RuntimeWarning)
                window = dist.cdf(dv_meta_from_conf_ub) - dist.cdf(dv_meta_from_conf_lb)
                likelihood = (dv_meta_from_conf > 1e-8) * window + \
                             (dv_meta_from_conf <= 1e-8) * dist.cdf(dv_meta_from_conf + binsize_pos)
            if final:
                try:
                    cdf2 = dist.cdf(dv_meta_from_conf.astype(maxfloat) + binsize_pos).astype(np.float64)
                except TypeError:
                    # may currently fail for the normal distribution because np.float128 is not supported
                    cdf2 = dist.cdf(dv_meta_from_conf + binsize_pos)
                likelihood_pdf = (dv_meta_from_conf > 1e-8) * \
                    dist.pdf(dv_meta_from_conf.astype(maxfloat)).astype(np.float64) + (dv_meta_from_conf <= 1e-8) * cdf2
            else:
                likelihood_pdf = None
        else:
            likelihood = dist.cdf(dv_meta_from_conf_ub) - dist.cdf(dv_meta_from_conf_lb)
            if final:
                likelihood_pdf = dist.pdf(dv_meta_from_conf)
            else:
                likelihood_pdf = None

        if not self.cfg.detection_model and not self.cfg.experimental_include_incongruent_dv:
            likelihood[self.model.dv_sens_considered_invalid] = np.nan

        return params_meta, noise_meta, dist, dv_meta_considered, likelihood, \
               likelihood_pdf, criteria_meta, levels_meta, punishment_factor  # noqa

    def _negll_meta_noisyreport(self, params, ignore_warnings=False, mock_binsize=None, final=False,
                                return_noise=False, return_criteria=False, return_levels=False,
                                constraint_mode=False):
        """
        Minimization function for the noisy-report model

        Parameters:
        -----------
        params : array-like of shape (nparams)
            Parameter array of the metacognitive level.
        ignore_warnings : bool
            If True, suppress warnings during minimization.
        mock_binsize : float
            Binsize around empirical confidence ratings to evaluate the likelihood. If not None, also returns the
            likelihood variable.
        final : bool
            If True, return latent variables and parameters.
        return_noise : bool
            If True, only return the metacognitive noise array (auxiliarly option to define constraints on the noise
            array).
        return_criteria : bool
            If True, only return confidence criteria (auxiliarly option to define constraints on confidence criteria).
        return_levels : bool
            If True, only return confidence levels (auxiliarly option to define constraints on confidence levels).
        constraint_mode : bool
            If True, method runs during scipy optimize constraint testing.

        Returns:
        --------
        By default, the method returns the negative log likelihood. However, depending on the arguments various
        combinations of variables are returned (see Parameters).
        negll: float
            Negative (summed) log likelihood.
        """

        params_meta, noise_meta, model_confidence, dist, dv_meta_considered, likelihood, likelihood_pdf, \
            punishment_factor = \
            self._helper_negll_meta_noisyreport(params, ignore_warnings, mock_binsize, final, return_noise,
                                                return_criteria, return_levels, constraint_mode)

        # compute weighted cumulative negative log likelihood
        likelihood_weighted_cum = np.nansum(likelihood * self.model.dv_sens_pmf, axis=1)
        if self.cfg.experimental_likelihood:
            # use an upper bound for the negative log likelihood based on a uniform 'guessing' model
            negll = min(self.model.max_negll, -np.sum(np.log(np.maximum(likelihood_weighted_cum, 1e-200))))
        else:
            negll = -np.sum(np.log(np.maximum(likelihood_weighted_cum, self.cfg.min_likelihood_meta)))
        negll *= punishment_factor

        if final:
            return negll, params_meta, noise_meta, likelihood, dv_meta_considered, likelihood_weighted_cum, \
                   likelihood_pdf  # noqa
        elif mock_binsize is not None:
            return negll, likelihood
        else:
            return negll

    def _helper_negll_meta_noisyreport(self, params, ignore_warnings=False, mock_binsize=None, final=False,
                                       return_noise=False, return_criteria=False, return_levels=False,
                                       constraint_mode=False):

        bl = self.cfg.paramset_meta.base_len
        params_meta = {p: params[int(np.sum(bl[:i]))] if n == 1 else [params[int(np.sum(bl[:i])) + j] for j in range(n)]
                       for i, (p, n) in enumerate(zip(self.cfg.paramset_meta.base_names, bl))}

        punishment_factor = 1

        if '_criteria' in self.cfg.meta_link_function:
            criteria_meta, levels_meta = self._prepare_criteria(params_meta)
            if return_criteria:
                return criteria_meta
            if return_levels:
                return levels_meta
            if not final:
                bds = [(params[i] >= b[0]) & (params[i] <= b[1]) for i, b in enumerate(self.cfg.paramset_meta.bounds)]
                cns = [con['fun'](params) >= 0 for con in self.cfg.paramset_meta.constraints]
                if not np.all(cns) or not np.all(bds):
                    # if not self._punish_message:
                    #     print('punish!')
                    #     self._punish_message = True
                    punishment_factor = 10
        else:
            criteria_meta, levels_meta = None, None

        dv_meta_considered = self._compute_dv_meta(params_meta)

        self.model.confidence = self._link_function(dv_meta_considered, params_meta, criteria_meta=criteria_meta,
                                                    levels_meta=levels_meta, constraint_mode=constraint_mode)

        noise_meta = self._noise_meta_transform(self.model.confidence, params_meta, ignore_warnings=ignore_warnings)
        if return_noise:
            return noise_meta
        if (self.cfg.meta_noise_dist == 'beta') and np.any(noise_meta > 0.5):
            if np.max(noise_meta) < 0.5 + 1e-5:
                noise_meta = np.minimum(0.5, noise_meta)
            else:
                warnings.warn(f'max(noise_intercept_meta) = {np.max(noise_meta):.2f}, but maximum allowed '
                              f'value for noise_intercept_meta is 0.5 for metacognitive type '
                              f'{self.cfg.meta_noise_type} and noise model {self.cfg.meta_noise_dist}')
                punishment_factor = np.max(noise_meta) / 0.5
                noise_meta = np.minimum(0.5, noise_meta)

        binsize = self.cfg.binsize_meta if mock_binsize is None else mock_binsize
        if self.cfg.experimental_wrap_binsize_meta:
            wrap_neg = binsize - np.abs(np.minimum(1, self.data.confidence_2d + binsize) - self.data.confidence_2d)
            wrap_pos = binsize - np.abs(np.maximum(0, self.data.confidence_2d - binsize) - self.data.confidence_2d)
            binsize_neg, binsize_pos = binsize + wrap_neg, binsize + wrap_pos
        else:
            binsize_neg, binsize_pos = binsize, binsize
        dist = get_dist(self.cfg.meta_noise_dist, mode=self.model.confidence, scale=noise_meta,
                        meta_noise_type='noisy_report', lookup_table=self.lookup_table)
        # compute the probability of the actual confidence ratings given the pred confidence
        if self.cfg.meta_noise_dist.startswith('censored_'):
            if self.cfg.meta_noise_dist.endswith('gumbel'):
                binsize_neg = np.array(binsize_neg).astype(maxfloat)
                binsize_pos = np.array(binsize_pos).astype(maxfloat)
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', RuntimeWarning)
                window = (dist.cdf(np.minimum(1, self.data.confidence_2d + binsize_pos)) -
                          dist.cdf(np.maximum(0, self.data.confidence_2d - binsize_neg)))
            likelihood = (((self.data.confidence_2d > 1e-8) & (self.data.confidence_2d < 1 - 1e-8)) * window +
                          (self.data.confidence_2d <= 1e-8) * dist.cdf(binsize_pos).astype(np.float64) +
                          (self.data.confidence_2d >= 1 - 1e-8) * (1 - dist.cdf(1 - binsize_neg))).astype(np.float64)
            if final:
                likelihood_pdf = \
                    (((self.data.confidence_2d > 1e-8) & (self.data.confidence_2d < 1 - 1e-8)) *
                        dist.pdf(self.data.confidence_2d.astype(maxfloat)) +
                     (self.data.confidence_2d <= 1e-8) * dist.cdf(binsize_pos) +
                        (self.data.confidence_2d >= 1 - 1e-8) * (1 - dist.cdf(1 - binsize_neg))).astype(np.float64)
            else:
                likelihood_pdf = None
        else:
            with warnings.catch_warnings():
                # catch this warning, which doesn't make any sense (output is valid if this happens)
                warnings.filterwarnings('ignore', category=RuntimeWarning, module='scipy.stats',
                                        message='divide by zero encountered in _beta_cdf')
                likelihood = dist.cdf(np.minimum(1, self.data.confidence_2d + binsize_pos)) - \
                             dist.cdf(np.maximum(0, self.data.confidence_2d - binsize_neg))  # noqa
            if final:
                likelihood_pdf = dist.pdf(self.data.confidence_2d)
            else:
                likelihood_pdf = None

        if not self.cfg.detection_model and not self.cfg.experimental_include_incongruent_dv:
            likelihood[self.model.dv_sens_considered_invalid] = np.nan
            self.model.confidence[self.model.dv_sens_considered_invalid] = np.nan

        return params_meta, noise_meta, self.model.confidence, dist, dv_meta_considered, likelihood, likelihood_pdf, \
            punishment_factor

    def _noise_sens_transform(self, stimuli=None, params_sens=None):
        """
        Helper function to call the sensory noise transformation function.
        """
        if stimuli is None:
            stimuli = self.model.stimuli_final
        if params_sens is None:
            params_sens = self.model.params_sens
        if self.cfg.enable_noise_sens:
            return noise_sens_transform(
                stimuli=stimuli, function_noise_transform_sens=self.cfg.function_noise_transform_sens, **params_sens
            )
        else:
            params_sens = {} if params_sens is None else params_sens
            return noise_sens_transform(
                stimuli=stimuli, function_noise_transform_sens=self.cfg.function_noise_transform_sens,
                **{**params_sens, 'noise_sens': self.cfg.noise_sens_default}
            )

    def _noise_meta_transform(self, confidence_or_dv_meta, params_meta, ignore_warnings=False):
        """
        Helper function to call the metacognitive noise transformation function.
        """
        if self.cfg.enable_noise_meta:
            noise_meta_transformed = noise_meta_transform(
                confidence_or_dv_meta=confidence_or_dv_meta, dv_sens=self.model.dv_sens_considered,
                function_noise_transform_meta=self.cfg.function_noise_transform_meta,
                ignore_warning=ignore_warnings,  # noqa
                **self.model.params_sens, **params_meta
            )
            return np.maximum(self.cfg.noise_meta_min, noise_meta_transformed)
        else:
            return self.cfg.noise_meta_default

    def _link_function(self, dv_meta, params_meta, criteria_meta=None, levels_meta=None, constraint_mode=False):
        """
        Helper function to call the link function.
        """
        return link_function(
            dv_meta=dv_meta, link_fun=self.cfg.meta_link_function, dv_sens=self.model.dv_sens_considered,
            stimuli=self.model.stimuli_final, function_noise_transform_sens=self.cfg.function_noise_transform_sens,
            nchannels=self.cfg.detection_model_nchannels, criteria_meta=criteria_meta, levels_meta=levels_meta,
            constraint_mode=constraint_mode,
            **self.model.params_sens, **params_meta
        )

    def _link_function_inv(self, confidence, link_fun, params_meta, criteria_meta=None, levels_meta=None):
        """
        Helper function to call the inverse link function.
        """
        return link_function_inv(
            confidence=confidence, link_fun=link_fun, stimuli=self.model.stimuli_final,
            dv_sens=self.model.dv_sens_considered, function_noise_transform_sens=self.cfg.function_noise_transform_sens,
            nchannels=self.cfg.detection_model_nchannels, criteria_meta=criteria_meta, levels_meta=levels_meta,
            **self.model.params_sens, **params_meta
        )

    def _compute_dv_sens(self):
        """
        Compute sensory decision values
        """

        noise_sens = self._noise_sens_transform()
        thresh_sens = _check_param(self.model.params_sens['thresh_sens'] if self.cfg.enable_thresh_sens else 0)
        bias_sens = _check_param(self.model.params_sens['bias_sens'] if self.cfg.enable_bias_sens else 0)

        cond_neg, cond_pos = (self.model.stimuli_final < 0).squeeze(), (self.model.stimuli_final >= 0).squeeze()
        self.model.stimuli_final_thresh = np.full(self.model.stimuli_final.shape, np.nan)
        # self.model.stimuli_final_thresh[cond_neg] = self.model.stimuli_final[cond_neg] - \
        #     np.sign(self.model.stimuli_final[cond_neg]) * thresh_sens[0]
        # self.model.stimuli_final_thresh[cond_pos] = self.model.stimuli_final[cond_pos] - \
        #     np.sign(self.model.stimuli_final[cond_pos]) * thresh_sens[1]
        self.model.stimuli_final_thresh[cond_neg] = self.model.stimuli_final[cond_neg]
        self.model.stimuli_final_thresh[cond_pos] = self.model.stimuli_final[cond_pos]
        dv_sens_mode = np.full(self.model.stimuli_final.shape, np.nan)
        dv_sens_mode[cond_neg] = (np.abs(self.model.stimuli_final[cond_neg]) >= thresh_sens[0]) * \
            self.model.stimuli_final_thresh[cond_neg] + bias_sens[0]
        dv_sens_mode[cond_pos] = (np.abs(self.model.stimuli_final[cond_pos]) >= thresh_sens[1]) * \
            self.model.stimuli_final_thresh[cond_pos] + bias_sens[1]

        # self.model.stimuli_final_thresh = self.model.stimuli_final - np.sign(self.model.stimuli_final) * thresh_sens
        # dv_sens_mode = (np.abs(self.model.stimuli_final) >= thresh_sens) * self.model.stimuli_final_thresh - bias_sens

        if self.cfg.detection_model:
            nchannels = self.cfg.detection_model_nchannels  # number of sensory channels
            p_active = np.full(dv_sens_mode.shape, np.nan)  # prob. channel active
            p_active[cond_neg] = np.tanh(np.abs(dv_sens_mode[cond_neg]) / noise_sens[cond_neg])
            p_active[cond_pos] = np.tanh(np.abs(dv_sens_mode[cond_pos]) / noise_sens[cond_pos])

            signs = np.sign(dv_sens_mode)
            dv_sens_mode = signs * p_active * nchannels
            self.model.dv_sens_considered = signs * np.tile(np.arange(0, nchannels + 1), (len(dv_sens_mode), 1))
            self.model.dv_sens_pmf = binom(nchannels, p_active).pmf(np.abs(self.model.dv_sens_considered))

            # this doesn't work, possibly because the beta distribution is zero at p=0?
            # (in contrast to the binomial distribution)
            # nactive = p_active * nchannels
            # ninactive = nchannels - nactive
            # p_active_range = np.linspace(1e-5, 1-1e-5, self.cfg.nbins_dv)
            # self.model.dv_sens_considered = signs * np.tile(p_active_range*nchannels, (len(dv_sens_mode), 1))
            # self.model.dv_sens_pmf = beta(nactive+1, ninactive+1).pdf(p_active_range)

            # signs = np.sign(dv_sens_mode)
            # dv_sens_range = np.linspace(0, np.abs(dv_sens_mode.flatten()) +
            #                             self.cfg.max_dv_deviation, int((self.cfg.nbins_dv))).T
            # self.model.dv_sens_considered = signs * dv_sens_range * noise_transform_sens
            # p_active_considered = np.tanh(np.abs(self.model.dv_sens_considered) / noise_transform_sens)
            # nactive_considered = p_active_considered * nchannels
            # self.model.dv_sens_pmf = binom(nchannels, p_active).pmf(np.round(nactive_considered).astype(int))

            # signs = np.sign(dv_sens_mode)
            # dv_sens_mode = signs * p_active * nchannels
            # nactive_considered = np.tile(np.arange(0, nchannels+1), (len(dv_sens_mode), 1))
            # self.model.dv_sens_considered = signs * noise_transform_sens * \
            #                                 np.arctanh(np.minimum(nchannels-1e-5, nactive_considered) / nchannels)
            # self.model.dv_sens_pmf = binom(nchannels, p_active).pmf(nactive_considered)

        else:
            range_ = np.linspace(0, self.cfg.max_dv_deviation, int((self.cfg.nbins_dv + 1) / 2))[1:]
            dv_sens_range = np.hstack((-range_[::-1], 0, range_))
            self.model.dv_sens_considered = np.full((dv_sens_mode.shape[0], dv_sens_range.shape[0]), np.nan)
            self.model.dv_sens_considered[cond_neg] = dv_sens_mode[cond_neg] + dv_sens_range * noise_sens[cond_neg]
            self.model.dv_sens_considered[cond_pos] = dv_sens_mode[cond_pos] + dv_sens_range * noise_sens[cond_pos]

            logistic_neg = logistic_dist(loc=dv_sens_mode[cond_neg], scale=noise_sens[cond_neg] * np.sqrt(3) / np.pi)
            logistic_pos = logistic_dist(loc=dv_sens_mode[cond_pos], scale=noise_sens[cond_pos] * np.sqrt(3) / np.pi)
            margin_neg = noise_sens[cond_neg] * self.cfg.max_dv_deviation / self.cfg.nbins_dv
            margin_pos = noise_sens[cond_pos] * self.cfg.max_dv_deviation / self.cfg.nbins_dv
            self.model.dv_sens_pmf = np.full(self.model.dv_sens_considered.shape, np.nan)
            self.model.dv_sens_pmf[cond_neg] = (logistic_neg.cdf(self.model.dv_sens_considered[cond_neg] + margin_neg) -
                                                logistic_neg.cdf(self.model.dv_sens_considered[cond_neg] - margin_neg))
            self.model.dv_sens_pmf[cond_pos] = (logistic_pos.cdf(self.model.dv_sens_considered[cond_pos] + margin_pos) -
                                                logistic_pos.cdf(self.model.dv_sens_considered[cond_pos] - margin_pos))
            # normalize PMF
            self.model.dv_sens_pmf = self.model.dv_sens_pmf / self.model.dv_sens_pmf.sum(axis=1).reshape(-1, 1)
            # invalidate invalid decision values
            if not self.cfg.experimental_include_incongruent_dv:
                self.model.dv_sens_considered_invalid = np.sign(self.model.dv_sens_considered) != \
                                                        np.sign(self.data.choices_2d - 0.5)
                self.model.dv_sens_pmf[self.model.dv_sens_considered_invalid] = np.nan

            if self.cfg.experimental_likelihood:
                # self.cfg.binsize_meta*2 is the probability for a given confidence rating assuming a uniform
                # distribution for confidence. This 'confidence guessing model' serves as a upper bound for the
                # negative log likelihood.
                min_likelihood = self.cfg.binsize_meta*2*np.ones(self.model.dv_sens_pmf.shape)
                min_likelihood_weighted_cum = np.nansum(min_likelihood * self.model.dv_sens_pmf, axis=1)
                self.model.max_negll = -np.log(min_likelihood_weighted_cum).sum()

        self.model.dv_sens_considered_abs = np.abs(self.model.dv_sens_considered)
        self.model.dv_sens_mode = dv_sens_mode.flatten()

    def _compute_dv_meta(self, params_meta):
        """
        Compute metacognitive decision values
        """
        if self.cfg.enable_evidence_bias_mult_meta == 1:
            dv_meta_considered = params_meta['evidence_bias_mult_meta'] * self.model.dv_sens_considered_abs
        elif self.cfg.enable_evidence_bias_mult_meta == 2:
            dv_meta_considered = np.full(self.model.dv_sens_considered.shape, np.nan)
            neg, pos = self.model.dv_sens_considered < 0, self.model.dv_sens_considered >= 0
            dv_meta_considered[neg] = params_meta['evidence_bias_mult_meta'][0] * self.model.dv_sens_considered_abs[neg]
            dv_meta_considered[pos] = params_meta['evidence_bias_mult_meta'][1] * self.model.dv_sens_considered_abs[pos]
        else:
            dv_meta_considered = self.model.dv_sens_considered_abs

        if self.cfg.enable_evidence_bias_add_meta == 1:
            dv_meta_considered = np.maximum(0, dv_meta_considered + params_meta['evidence_bias_add_meta'])
        elif self.cfg.enable_evidence_bias_add_meta == 2:
            dv_meta_considered = np.full(self.model.dv_sens_considered.shape, np.nan)
            neg, pos = self.model.dv_sens_considered < 0, self.model.dv_sens_considered >= 0
            dv_meta_considered[neg] = np.maximum(0, dv_meta_considered[neg] + params_meta['evidence_bias_add_meta'][0])
            dv_meta_considered[pos] = np.maximum(0, dv_meta_considered[pos] + params_meta['evidence_bias_add_meta'][1])
        return dv_meta_considered

    def _prepare_criteria(self, params_meta):
        ncrit_meta = int(self.cfg.meta_link_function.split('_')[0])

        if self.cfg.enable_criteria_meta == 2:
            criteria_meta = [[params_meta[f'criterion{i}_meta'][j] for i in range(ncrit_meta)] for j in range(2)]
        else:
            criteria_meta = [params_meta[f'criterion{i}_meta'] for i in range(ncrit_meta)]

        if self.cfg.enable_levels_meta == 2:
            levels_meta = [[params_meta[f'level{i}_meta'][j] for i in range(ncrit_meta)] for j in range(2)]
        elif self.cfg.enable_levels_meta == 1:
            levels_meta = [params_meta[f'level{i}_meta'] for i in range(ncrit_meta)]
        else:
            levels_meta = None

        return criteria_meta, levels_meta

    def _check_fit(self):
        if not self.sens_is_fitted and not self.meta_is_fitted:
            raise RuntimeError('Please fit the model before plotting.')
        elif self.sens_is_fitted and not self.meta_is_fitted:
            raise RuntimeError('Only the sensory level was fitted. Please also fit the metacognitive level to plot'
                               'a link function.')

    def plot_link_function(self, **kwargs):
        self._check_fit()
        plot_link_function(
            self.data.stimuli, self.data.confidence, self.model.dv_sens_mode, self.model.params, cfg=self.cfg, **kwargs
        )

    def plot_confidence_dist(self, **kwargs):
        self._check_fit()
        varlik = self.model.dv_meta_considered if self.cfg.meta_noise_type == 'noisy_readout' else self.model.confidence
        plot_confidence_dist(
            self.cfg, self.data.stimuli, self.data.confidence, self.model.params, var_likelihood=varlik,
            noise_meta_transformed=self.model.noise_meta, dv_sens=self.model.dv_sens_considered,
            likelihood_weighting=self.model.dv_sens_pmf, **kwargs
        )


def load_dataset(name, verbose=True, return_params=False, return_dv_sens=False, return_cfg=False):
    import gzip
    path = os.path.join(pathlib.Path(__file__).parent.resolve(), 'demo_data', f'example_data_{name}.pkl.gz')
    if os.path.exists(path):
        with gzip.open(path, 'rb') as f:
            stimuli, choices, confidence, params, cfg, dv_sens, stats = pickle.load(f)
    else:
        raise FileNotFoundError(f'[Dataset does not exist!] No such file: {path}')

    if verbose:
        print(f"Loading dataset '{name}' which was generated as follows:")
        print('..Generative model:')
        print(f'{TAB}Metatacognitive noise type: {cfg.meta_noise_type}')
        print(f'{TAB}Metatacognitive noise distribution: {cfg.meta_noise_dist}')
        print(f'{TAB}Link function: {cfg.meta_link_function}')
        print('..Generative parameters:')
        for i, (k, v) in enumerate(params.items()):
            if hasattr(v, '__len__'):
                print(f"{TAB}{k}: {[float(f'{x:.4g}') for x in v]}")
            else:
                print(f'{TAB}{k}: {v:.4g}')
        print('..Characteristics:')
        print(f'{TAB}No. subjects: {1 if stimuli.ndim == 1 else len(stimuli)}')
        print(f'{TAB}No. samples: {stimuli.shape[0] if stimuli.ndim == 1 else stimuli.shape[1]}')
        print(f"{TAB}Type 1 performance: {100*stats['performance']:.1f}%")
        if not cfg.skip_meta:
            print(f"{TAB}Avg. confidence: {stats['confidence']:.3f}")
            print(f"{TAB}M-Ratio: {stats['mratio']:.3f}")

    return_list = [stimuli, choices, confidence]
    if return_params:
        return_list += [params]
    if return_cfg:
        return_list += [cfg]
    if return_dv_sens:
        return_list += [dv_sens]
    return tuple(return_list)



import warnings

import numpy as np
from scipy.stats import norm, lognorm, beta, betaprime, gamma, uniform, gumbel_r

from .util import maxfloat
from .fast_truncnorm import truncnorm

META_NOISE_DIST = (
    'beta', 'beta_std',
    'norm', 'gumbel', 'lognorm', 'lognorm_varstd', 'betaprime', 'gamma',
    'censored_norm', 'censored_gumbel', 'censored_lognorm', 'censored_lognorm_varstd',
    'censored_betaprime', 'censored_gamma',
    'truncated_norm', 'truncated_norm_lookup', 'truncated_norm_fit',
    'truncated_gumbel', 'truncated_gumbel_lookup',
    'truncated_lognorm', 'truncated_lognorm_varstd'
)
META_NOISE_DIST_REPORT_ONLY = ('beta', 'beta_std')
META_NOISE_DIST_READOUT_ONLY = ('lognorm', 'lognorm_varstd', 'betaprime', 'gamma')


def _lognorm_params(mode, stddev):
    """
    Compute scipy lognorm's shape and scale for a given mode and SD
    The analytical formula is exact and was computed with WolframAlpha.

    Parameters
    ----------
    mode : float or array-like
        Mode of the distribution.
    stddev : float or array-like
        Standard deviation of the distribution.

    Returns
    ----------
    shape : float
        Scipy lognorm shape parameter.
    scale : float
        Scipy lognorm scale parameter.
    """
    mode = np.maximum(1e-5, mode)
    a = stddev ** 2 / mode ** 2
    x = 1 / 4 * np.sqrt(np.maximum(1e-300, -(16 * (2 / 3) ** (1 / 3) * a) / (
                np.sqrt(3) * np.sqrt(256 * a ** 3 + 27 * a ** 2) - 9 * a) ** (1 / 3) +
                                   2 * (2 / 3) ** (2 / 3) * (
                                               np.sqrt(3) * np.sqrt(256 * a ** 3 + 27 * a ** 2) - 9 * a) ** (
                                               1 / 3) + 1)) + \
        1 / 2 * np.sqrt(
        (4 * (2 / 3) ** (1 / 3) * a) / (np.sqrt(3) * np.sqrt(256 * a ** 3 + 27 * a ** 2) - 9 * a) ** (1 / 3) -
        (np.sqrt(3) * np.sqrt(256 * a ** 3 + 27 * a ** 2) - 9 * a) ** (1 / 3) / (2 ** (1 / 3) * 3 ** (2 / 3)) +
        1 / (2 * np.sqrt(np.maximum(1e-300, -(16 * (2 / 3) ** (1 / 3) * a) / (
                    np.sqrt(3) * np.sqrt(256 * a ** 3 + 27 * a ** 2) - 9 * a) ** (1 / 3) +
                                    2 * (2 / 3) ** (2 / 3) * (
                                                np.sqrt(3) * np.sqrt(256 * a ** 3 + 27 * a ** 2) - 9 * a) ** (
                                                1 / 3) + 1))) + 1 / 2) + \
        1 / 4
    shape = np.sqrt(np.log(x))
    scale = mode * x
    return shape, scale


class truncated_lognorm:  # noqa
    """
    Implementation of the truncated lognormal distribution.
    Only the upper truncation bound is supported as the lognormal distribution is naturally lower-bounded at zero.

    Parameters
    ----------
    loc : float or array-like
        Scipy lognorm's loc parameter.
    scale : float or array-like
        Scipy lognorm's scale parameter.
    s : float or array-like
        Scipy lognorm's s parameter.
    b : float or array-like
        Upper truncation bound.
    """
    def __init__(self, loc, scale, s, b):
        self.loc = loc
        self.scale = scale
        self.s = s
        self.b = b
        self.dist = lognorm(loc=loc, scale=scale, s=s)
        self.lncdf_b = self.dist.cdf(self.b)

    def pdf(self, x):
        pdens = (x <= self.b) * self.dist.pdf(x) / self.lncdf_b
        return pdens

    def cdf(self, x):
        cdens = (x > self.b) + (x <= self.b) * self.dist.cdf(x) / self.lncdf_b
        return cdens

    def rvs(self, size=None):
        if size is None:
            if hasattr(self.scale, '__len__'):
                size = self.scale.shape
            else:
                size = 1
        cdens = uniform(loc=0, scale=self.b).rvs(size)
        x = self.cdf_inv(cdens)
        return x

    def cdf_inv(self, cdens):
        x = (cdens >= 1) * self.b + (cdens < 1) * self.dist.ppf(cdens * self.lncdf_b)
        return x


class truncated_gumbel:  # noqa
    """
    Implementation of the truncated Gumbel distribution.

    Parameters
    ----------
    loc : float or array-like
        Scipy gumbel_r's loc parameter.
    scale : float or array-like
        Scipy gumbel_r's scale parameter.
    a : float or array-like
        Lower truncation bound.
    b : float or array-like
        Upper truncation bound.
    """
    def __init__(self, loc, scale, a=0, b=np.inf):
        self.loc = loc
        self.scale = scale
        self.a = a
        self.b = b
        self.dist = gumbel_r(loc=loc, scale=scale)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            self.cdf_to_a = self.dist.cdf(self.a)
        self.cdf_a_to_b = self.dist.cdf(self.b) - self.cdf_to_a

    def pdf(self, x):
        pdens = ((x > self.a) & (x < self.b)) * self.dist.pdf(maxfloat(x)).astype(np.float64) / self.cdf_a_to_b
        return pdens

    def cdf(self, x):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            cdf_to_x = self.dist.cdf(x)
        cdens = (x >= self.b) + ((x > self.a) & (x < self.b)) * (cdf_to_x - self.cdf_to_a) / self.cdf_a_to_b
        return cdens

    def rvs(self, size=None):
        if size is None:
            if hasattr(self.scale, '__len__'):
                size = self.loc.shape
            else:
                size = 1
        cdens = uniform(loc=self.cdf_to_a, scale=self.cdf_a_to_b).rvs(size)
        x = self.cdf_inv(cdens)
        return x

    def cdf_inv(self, cdens):
        x = (cdens > self.cdf_to_a) * self.dist.ppf(cdens)
        return x


def get_dist(meta_noise_dist, mode, scale, meta_noise_type='noisy_report', lookup_table=None):
    """
    Helper function to select appropriately parameterized metacognitive noise distributions.
    """

    if meta_noise_dist not in META_NOISE_DIST:
        raise ValueError(f"Unkonwn distribution '{meta_noise_dist}'.")
    elif (meta_noise_type == 'noisy_report') and meta_noise_dist in META_NOISE_DIST_READOUT_ONLY:
        raise ValueError(f"Distribution '{meta_noise_dist}' is only valid for noisy-readout models.")
    elif (meta_noise_type == 'noisy_readout') and meta_noise_dist in META_NOISE_DIST_REPORT_ONLY:
        raise ValueError(f"Distribution '{meta_noise_dist}' is only valid for noisy-report models.")

    if meta_noise_dist.startswith('censored_'):
        distname = meta_noise_dist[meta_noise_dist.find('_') + 1:]
    else:
        distname = meta_noise_dist

    if distname == 'norm':
        dist = norm(loc=mode, scale=scale)
    elif distname == 'gumbel':
        dist = gumbel_r(loc=mode, scale=scale * np.sqrt(6) / np.pi)
    elif distname == 'lognorm_varstd':
        dist = lognorm(loc=0, scale=np.maximum(1e-5, mode) * np.exp(scale ** 2), s=scale)
    elif distname == 'lognorm':
        shape, scale = _lognorm_params(np.maximum(1e-5, mode), scale)
        dist = lognorm(loc=0, scale=scale, s=shape)
    elif distname == 'lognorm_varstd':
        dist = lognorm(loc=0, scale=np.maximum(1e-5, mode) * np.exp(scale ** 2), s=scale)
    elif distname == 'beta':
        a = mode * (1 / scale - 2) + 1
        b = (1 - mode) * (1 / scale - 2) + 1
        dist = beta(a, b)
    elif distname == 'beta_std':
        mode = np.maximum(1e-5, np.minimum(1-1e-5, mode))
        a = 1 + (1 - mode) * mode**2 / scale**2
        b = (1/mode - 1) * a - 1/mode + 2
        dist = beta(a, b)
    elif distname == 'betaprime':
        mode = np.minimum(1 / scale - 1 - 1e-3, mode)
        a = (mode * (1 / scale + 1) + 1) / (mode + 1)
        b = (1 / scale - mode - 1) / (mode + 1)
        dist = betaprime(a, b)
    elif distname == 'gamma':
        a = (mode + np.sqrt(mode**2 + 4*scale**2))**2 / (4*scale**2)
        b = (mode + np.sqrt(mode**2 + 4*scale**2)) / (2*scale**2)
        dist = gamma(a=a, loc=0, scale=1/b)
    elif distname.startswith('truncated_'):
        if meta_noise_type == 'noisy_report':
            if distname.endswith('_lookup') and (distname.startswith('truncated_norm_') or
                                                 distname.startswith('truncated_gumbel_')):
                m_ind = np.searchsorted(lookup_table['mode'], mode)
                scale = lookup_table['scale'][np.abs(lookup_table['truncscale'][m_ind] - scale).argmin(axis=-1)]
            elif distname == 'truncated_norm_fit':
                mode_ = mode.copy()
                mode_[mode_ < 0.5] = 1 - mode_[mode_ < 0.5]
                scale = np.minimum(scale, 1/np.sqrt(12) - 1e-3)
                alpha1, beta1, alpha2, beta2, theta = -0.1512684, 4.15388204, -1.01723445, 2.47820677, 0.73799941
                scale = scale / (beta1*mode_**alpha1*np.sqrt(1/12-scale**2)) * (mode_ < theta) + \
                    (np.arctanh(2*np.sqrt(3)*scale) / (beta2*mode_**alpha2)) * (mode_ >= theta)
            if distname.startswith('truncated_norm'):
                dist = truncnorm(-mode / scale, (1 - mode) / scale, loc=mode, scale=scale)
            elif distname.startswith('truncated_gumbel'):
                dist = truncated_gumbel(loc=mode, scale=scale * np.sqrt(6) / np.pi, b=1)
            elif distname == 'truncated_lognorm':
                shape, scale = _lognorm_params(np.maximum(1e-5, mode), scale)
                dist = truncated_lognorm(loc=0, scale=scale, s=shape, b=1)
            elif distname == 'truncated_lognorm_varstd':
                dist = truncated_lognorm(loc=0, scale=np.maximum(1e-5, mode) * np.exp(scale ** 2), s=scale, b=1)
        elif meta_noise_type == 'noisy_readout':
            if distname.endswith('_lookup') and (distname.startswith('truncated_norm_') or
                                                 distname.startswith('truncated_gumbel_')):
                m_ind = np.searchsorted(lookup_table['mode'], np.minimum(10, mode))  # atm 10 is the maximum mode
                scale = lookup_table['scale'][np.abs(lookup_table['truncscale'][m_ind] - scale).argmin(axis=-1)]
            elif distname == 'truncated_norm_fit':
                a, b, c, d, e, f = 0.88632051, -1.45129289, 0.25329918, 2.09066054, -1.2262868, 1.47179606
                scale = a*scale*(mode+1)**b + c*(mode+1)**f*(np.exp(np.minimum(100, d*scale*(mode+1)**e))-1)
            if distname.startswith('truncated_norm'):
                dist = truncnorm(-mode / scale, np.inf, loc=mode, scale=scale)
            elif distname.startswith('truncated_gumbel'):
                dist = truncated_gumbel(loc=mode, scale=scale * np.sqrt(6) / np.pi)
        else:
            raise ValueError(f"'{meta_noise_type}' is an unknown metacognitive type")

    return dist  # noqa


def get_dist_mean(distname, mode, scale):
    """
    Helper function to get the distribution mean of certain distributions.
    """
    if distname == 'gumbel':
        mean = mode + np.euler_gamma * scale * np.sqrt(6) / np.pi
    elif distname == 'norm':
        mean = mode
    else:
        raise ValueError(f"Distribution {distname} not supported.")
    return mean


def get_pdf(x, distname, mode, scale, lb=0, ub=1, meta_noise_type='noisy_report'):
    """
    Helper function to get the probability density of a distribution.
    """
    if distname.startswith('censored_'):
        likelihood_dist = distname[distname.find('_') + 1:]
        dist = get_dist(likelihood_dist, mode=mode, scale=scale)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            cdf_margin_low = dist.cdf(lb)
            cdf_margin_high = (1 - dist.cdf(ub))
        pdf = ((x > lb) & (x < ub)) * dist.pdf(x) + \
              (x <= lb) * cdf_margin_low + \
              (x >= ub) * cdf_margin_high
    else:
        pdf = get_dist(distname, mode, scale, meta_noise_type).pdf(x)
    return pdf


def get_likelihood(x, distname, mode, scale, lb=1e-8, ub=1 - 1e-8, binsize_meta=1e-3, logarithm=False):
    """
    Helper function to get the likelihood of a distribution.
    """
    if distname.startswith('censored_'):
        likelihood_dist = distname[distname.find('_') + 1:]
        dist = get_dist(likelihood_dist, mode=mode, scale=scale)
        if distname.endswith('gumbel'):
            x = x.astype(maxfloat)
        window = (dist.cdf(np.minimum(1, x + binsize_meta)) -
                  dist.cdf(np.maximum(0, x - binsize_meta))).astype(np.float64)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            cdf_margin_low = dist.cdf(binsize_meta)
            cdf_margin_high = (1 - dist.cdf(1 - binsize_meta))
        likelihood = ((x > lb) & (x < ub)) * window + \
                     (x <= lb) * cdf_margin_low + \
                     (x >= ub) * cdf_margin_high
    else:
        dist = get_dist(distname, mode, scale)
        likelihood = dist.cdf(x + binsize_meta) - dist.cdf(x - binsize_meta)
    return np.log(np.maximum(1e-4, likelihood)) if logarithm else likelihood



from dataclasses import make_dataclass

import numpy as np
from scipy.optimize import OptimizeResult

from .util import TAB, ReprMixin, spearman2d, pearson2d


class Parameter(ReprMixin):
    def __init__(self, guess, bounds, grid_range=None):
        """
        Class that defines the fitting characteristics of a Parameter.

        Parameters
        ----------
        guess : float
            Initial guess for the parameter value.
        bounds: array-like of length 2
            Parameter bounds. The first and second element indicate the lower and upper bound of the parameter.
        grid_range: array-like
            Points to visit in the initial parameter gridsearch search. It makes sense to restrict this to a range of
            _likely_ values, rather than minimal and maximal bounds.
        """
        self.guess = guess
        self.bounds = bounds
        self.grid_range = bounds if grid_range is None else grid_range

    def copy(self):
        return Parameter(self.guess, self.bounds, self.grid_range)


class ParameterSet(ReprMixin):
    def __init__(self, parameters, names, constraints=None):
        """
        Container class for all Parameters of a model.

        Parameters
        ----------
        parameters : dict[str, Parameter]
            The dictionary must have the form {parameter_name1: Parameter(..), parameter_name2: Parameter(..), ..}
        names: List[str]
            List of parameter names of a model.
        constraints: List[Dict]
            List of scipy minimize constraints (each constraint is a dictionary with keys 'type' and 'fun'}
        """
        self.base_names = names
        self.base_is_multiple = [isinstance(parameters[p], list) for p in names]
        self.base_len = [len(parameters[p]) if isinstance(parameters[p], list) else 1 for p in names]  # noqa
        self.names = sum([[f'{p}_{i}' for i in range(len(parameters[p]))] if isinstance(parameters[p], list)  # noqa
                          else [p] for p in names], [])
        self.guess = np.array(sum([[parameters[p][i].guess for i in range(len(parameters[p]))] if  # noqa
                                   isinstance(parameters[p], list) else [parameters[p].guess] for p in names], []))
        self.bounds = np.array(sum([[parameters[p][i].bounds for i in range(len(parameters[p]))] if  # noqa
                                   isinstance(parameters[p], list) else [parameters[p].bounds] for p in names], []))
        self.grid_range = np.array(sum([[parameters[p][i].grid_range for i in range(len(parameters[p]))] if  # noqa
                                   isinstance(parameters[p], list) else [parameters[p].grid_range] for p in names], []),
                                   dtype=object)
        self.constraints = [] if constraints is None else constraints
        self.nparams = len(names)


class Data(ReprMixin):

    def __init__(self, cfg, stimuli, choices, confidence):
        """
        Container class for behavioral data.

        Parameters
        ----------
        cfg : configuration.Configuration
            Settings
        stimuli : array-like of shape (n_samples)
            Array of signed stimulus intensity values, where the sign codes the stimulus category (+: cat1; -: cat2) and
            the absolut value codes the intensity. The scale of the data is not relevant, as a normalisation to [-1; 1]
            is applied.
        choices : array-like of shape (n_samples)
            Array of choices coded as 0 (cat1) and 1 (cat2) for the two stimulus categories. See parameter 'stimuli'
            for the definition of cat1 and cat2.
        confidence : array-like of shape (n_samples)
            Confidence ratings; must be normalized to the range [0;1].
        """

        self.cfg = cfg
        self.stimuli_unnorm = np.array(stimuli)
        self.choices = np.array(choices)
        self.confidence = np.array(confidence)

        self.stimuli = None
        self.stimulus_ids = None
        self.stimuli_unnorm_max = None
        self.stimuli_min = None
        self.correct = None
        self.stimuli_2d = None
        self.choices_2d = None
        self.confidence_2d = None
        self.nsamples = len(self.stimuli_unnorm)

    def preproc(self):

        self.stimulus_ids = (np.sign(self.stimuli_unnorm) == 1).astype(int)
        self.correct = self.stimulus_ids == self.choices

        # convert to 0/1 scheme if choices are provides as -1's and 1's
        if np.array_equal(np.unique(self.choices[~np.isnan(self.choices)]), [-1, 1]):
            self.choices[self.choices == -1] = 0

        self.stimuli_unnorm_max = np.max(np.abs(self.stimuli_unnorm))
        # Normalize stimuli
        if self.cfg.normalize_stimuli_by_max:
            self.stimuli = self.stimuli_unnorm / self.stimuli_unnorm_max
        else:
            if np.max(np.abs(self.stimuli_unnorm)) > 1:
                raise ValueError('Stimuli are not normalized to the range [-1; 1].')
            self.stimuli = self.stimuli_unnorm

        self.stimuli_min = np.abs(self.stimuli).min()

        if self.cfg.confidence_bounds_error > 0:
            self.confidence[self.confidence <= self.cfg.confidence_bounds_error] = 0
            self.confidence[self.confidence >= 1 - self.cfg.confidence_bounds_error] = 1

        self.stimuli_2d = self.stimuli.reshape(-1, 1)
        self.choices_2d = self.choices.reshape(-1, 1)
        self.confidence_2d = self.confidence.reshape(-1, 1)

    def summary(self, full=False):
        desc = dict(
            nsamples=self.nsamples
        )
        if full:
            dict_extended = dict(
                stimulus_ids=(self.stimuli >= 0).astype(int),
                stimuli_unnorm=self.stimuli_unnorm,
                stimuli_norm=self.stimuli,
                choices=self.choices,
                confidence=self.confidence
            )
            data_extended = make_dataclass('DataExtended', dict_extended.keys())
            data_extended.__module__ = '__main__'
            desc.update(dict(data=data_extended(**dict_extended)))
        data_summary = make_dataclass('DataSummary', desc.keys())

        def _repr(self_):
            txt = f'{self_.__class__.__name__}\n'
            txt += '\n'.join([f"\t{k}: {'<not displayed>' if k == 'data' else v}"
                              for k, v in self_.__dict__.items()])
            return txt

        data_summary.__repr__ = _repr
        data_summary.__module__ = '__main__'
        return data_summary(**desc)


class Model(ReprMixin):
    def __init__(self, cfg):
        """
        Container class for behavioral data.

        Parameters
        ----------
        cfg : configuration.Configuration
            Settings
        """
        self.cfg = cfg

        self.super_thresh = None
        self.stimuli_final = None
        self.stimuli_warped_tresh = None
        self.stimuli_warped_super = None
        self.choiceprob = None
        self.posterior = None
        self.dv_sens_considered = None
        self.dv_sens_considered_abs = None
        self.dv_sens_considered_invalid = None
        self.dv_meta_considered = None
        self.dv_sens_mode = None
        self.dv_sens_pmf = None
        self.dv_sens_pmf_renorm = None
        self.dv_meta_mode = None
        self.confidence = None
        self.nsamples = None

        self.likelihood_meta = None
        self.likelihood_meta_mode = None
        self.likelihood_meta_weighted_cum = None
        self.likelihood_meta_weighted_cum_renorm = None
        self.max_negll = None
        self.noise_meta = None

        self.params = None

        self.params_sens = None
        self.params_sens_full = None
        self.params_sens_unnorm = None

        self.params_meta = None
        self.likelihood_dist_meta = None

        self.fit = ModelFit()

    def store_sens(self, negll, params_sens, choiceprob, posterior, stimuli_final, stimulus_norm_coefficent):
        self.params_sens = params_sens
        self.stimuli_final = stimuli_final.reshape(-1, 1)
        self.choiceprob = choiceprob
        self.posterior = posterior
        self.fit.fit_sens.negll = negll
        self.nsamples = len(self.stimuli_final)
        if not self.cfg.enable_noise_sens:
            self.params_sens['noise_sens'] = self.cfg.noise_sens_default
        self.params_sens_unnorm = {k: list(np.array(v) * stimulus_norm_coefficent) if hasattr(v, '__len__') else
                                   v * stimulus_norm_coefficent for k, v in params_sens.items()}

    def store_meta(self, negll, params_meta, noise_meta, likelihood, dv_meta_considered, likelihood_weighted_cum,
                   likelihood_pdf):
        self.params_meta = params_meta
        self.noise_meta = noise_meta
        self.likelihood_meta = likelihood
        self.dv_meta_considered = dv_meta_considered
        if self.cfg.detection_model:
            col_ind = np.round(np.abs(self.dv_sens_mode)).astype(int)
            self.dv_meta_mode = dv_meta_considered[np.arange(len(dv_meta_considered)), col_ind]
            self.likelihood_meta_mode = likelihood[np.arange(len(likelihood)), col_ind]
        else:
            self.dv_meta_mode = dv_meta_considered[:, int((dv_meta_considered.shape[1] - 1) / 2)]
            self.likelihood_meta_mode = likelihood[:, int((likelihood.shape[1] - 1) / 2)]
        self.likelihood_meta_weighted_cum = likelihood_weighted_cum
        self.dv_sens_pmf_renorm = self.dv_sens_pmf / np.nansum(self.dv_sens_pmf, axis=1).reshape(-1, 1)
        self.likelihood_meta_weighted_cum_renorm = np.nansum(likelihood * self.dv_sens_pmf_renorm, axis=1)
        self.fit.fit_meta.negll = negll
        self.fit.fit_meta.negll_pdf = -np.sum(np.log(np.maximum(np.nansum(self.dv_sens_pmf *
                                                                          likelihood_pdf, axis=1), 1e-10)))

    def report_fit_sens(self, verbose=True):
        if verbose:
            for k, v in self.params_sens.items():
                true_string = '' if self.cfg.true_params is None else \
                    (f" (true: [{', '.join([f'{p:.3g}' for p in self.cfg.true_params[k]])}])" if  # noqa
                     hasattr(self.cfg.true_params[k], '__len__') else f' (true: {self.cfg.true_params[k]:.3g})')  # noqa
                value_string = f"[{', '.join([f'{p:.3g}' for p in v])}]" if hasattr(v, '__len__') else f'{v:.3g}'
                print(f'{TAB}[final] {k}: {value_string}{true_string}')
            # if hasattr(self.fit.fit_sens, 'execution_time'):
            #     print(f'Final stats: {self.fit.fit_sens.execution_time:.2g} secs, {self.fit.fit_sens.nfev} fevs')
            print(f'Final neg. LL: {self.fit.fit_sens.negll:.2f}')
            if self.cfg.true_params is not None and hasattr(self.fit.fit_sens, 'negll_true'):
                print(f'Neg. LL using true params: {self.fit.fit_sens.negll_true:.2f}')
            if hasattr(self.fit.fit_sens, 'execution_time'):
                print(f"Total fitting time: {self.fit.fit_sens.execution_time:.2g} secs")

    def report_fit_meta(self, verbose=True):
        if self.cfg.true_params is not None:
            if 'criteria' in self.cfg.meta_link_function:
                for i, k in enumerate(['', '_neg', '_pos']):
                    if f'criteria{k}_meta' in self.cfg.true_params:
                        if self.cfg.enable_levels_meta:
                            self.cfg.true_params.update(
                                {f"{'confidence_level' if np.mod(i, 2) else 'criterion'}_meta_{int(i / 2) + 1}{k}": v
                                 for i, v in
                                 enumerate(self.cfg.true_params[f'criteria{k}_meta'])})
                        else:
                            self.cfg.true_params.update({f"criterion_meta_{i + 1}{k}": v for i, v in
                                                         enumerate(self.cfg.true_params[f'criteria{k}_meta'])})

        if verbose:
            for k, v in self.params_meta.items():
                true_string = '' if self.cfg.true_params is None else \
                    (f" (true: [{', '.join([f'{p:.3g}' for p in self.cfg.true_params[k]])}])" if
                     hasattr(self.cfg.true_params[k], '__len__') else f' (true: {self.cfg.true_params[k]:.3g})')
                value_string = f"[{', '.join([f'{p:.3g}' for p in v])}]" if hasattr(v, '__len__') else f'{v:.3g}'
                print(f'{TAB}[final] {k}: {value_string}{true_string}')
            # if hasattr(self.fit.fit_meta, 'execution_time'):
                # print(f'Final stats: {self.fit.fit_meta.execution_time:.2g} secs, '
                #       f'{self.fit.fit_meta.nfev} fevs')
            print(f'Final neg. LL: {self.fit.fit_meta.negll:.2f}')
            if self.cfg.true_params is not None:
                if verbose:
                    print(f'Neg. LL using true params: {self.fit.fit_meta.negll_true:.2f}')
            if hasattr(self.fit.fit_meta, 'execution_time'):
                print(f"Total fitting time: {self.fit.fit_meta.execution_time:.2g} secs")

    def summary(self, extended=False, fun_meta=None, confidence_gen=None, confidence_emp=None):

        if not self.cfg.skip_meta:
            if self.cfg.detection_model:
                confidence_mode = self.confidence[np.arange(len(self.confidence)),
                                                  np.round(np.abs(self.dv_sens_mode)).astype(int)]
            else:
                confidence_mode = self.confidence[:, int((self.confidence.shape[1] - 1) / 2)]

            if confidence_gen is not None:
                confidence_tiled = np.tile(confidence_emp, (confidence_gen.shape[0], 1))
                self.fit.fit_meta.confidence_gen_pearson = \
                    np.tanh(np.nanmean(np.arctanh(pearson2d(confidence_gen, confidence_tiled))))
                self.fit.fit_meta.confidence_gen_spearman = \
                    np.tanh(np.nanmean(np.arctanh(
                        spearman2d(confidence_gen, confidence_tiled, axis=1))))
                self.fit.fit_meta.confidence_gen_mae = np.nanmean(np.abs(confidence_gen - confidence_emp))
                self.fit.fit_meta.confidence_gen_medae = np.nanmedian(np.abs(confidence_gen - confidence_emp))
            self.fit.fit_meta.negll_persample = self.fit.fit_meta.negll / self.nsamples
            self.fit.fit_meta.negll_prenoise = -np.nansum(np.log(np.maximum(self.likelihood_meta_mode, 1e-10)))
            self.fit.negll = self.fit.fit_sens.negll + self.fit.fit_meta.negll

        desc = dict(
            nsamples=self.nsamples,
            nparams_sens=self.cfg.paramset_sens.nparams,
            params_sens=self.params_sens,
            evidence_sens=dict(
                negll=self.fit.fit_sens.negll,
                aic=2*self.cfg.paramset_sens.nparams + 2*self.fit.fit_sens.negll,
                bic=2*np.log(self.nsamples) + 2*self.fit.fit_sens.negll
            ),
            params=self.params_sens,
            params_sens_unnorm=self.params_sens_unnorm,
            fit=self.fit
        )

        if self.cfg.true_params is not None:
            desc['evidence_sens'].update(
                negll_true=self.fit.fit_sens.negll_true,
                aic_true=2*self.cfg.paramset_sens.nparams + 2*self.fit.fit_sens.negll_true,
                bic_true=2*np.log(self.nsamples) + 2*self.fit.fit_sens.negll_true
            )

        if not self.cfg.skip_meta:
            desc.update(dict(
                nparams_meta=self.cfg.paramset_meta.nparams,
                params_meta=self.params_meta,
                params={**self.params_sens, **self.params_meta},
                nparams=self.cfg.paramset_sens.nparams + self.cfg.paramset_meta.nparams,
                evidence_meta=dict(
                    negll=self.fit.fit_meta.negll,
                    aic=2*self.cfg.paramset_meta.nparams + 2*self.fit.fit_meta.negll,
                    bic=2*np.log(self.nsamples) + 2*self.fit.fit_meta.negll
                )
            ))
            if self.cfg.true_params is not None:
                desc['evidence_meta'].update(
                    negll_true=self.fit.fit_meta.negll_true,
                    aic_true=2*self.cfg.paramset_meta.nparams + 2*self.fit.fit_meta.negll_true,
                    bic_true=2*np.log(self.nsamples) + 2*self.fit.fit_meta.negll_true
                )

            if extended:
                likelihood_01 = fun_meta(self.fit.fit_meta.x, mock_binsize=0.1)[1]
                likelihood_025 = fun_meta(self.fit.fit_meta.x, mock_binsize=0.25)[1]
                dict_extended = dict(
                    noise_meta=self.noise_meta,
                    likelihood=self.likelihood_meta,
                    likelihood_prenoise=self.likelihood_meta_mode,
                    likelihood_weighted_cum=self.likelihood_meta_weighted_cum,
                    likelihood_weighted_cum_renorm_01=np.nansum(likelihood_01 * self.dv_sens_pmf_renorm, axis=1),
                    likelihood_weighted_cum_renorm_025=np.nansum(likelihood_025 * self.dv_sens_pmf_renorm, axis=1),
                    confidence=self.confidence,
                    dv_meta=self.dv_meta_considered,
                    dv_sens=self.dv_sens_considered,
                    dv_sens_pmf=self.dv_sens_pmf,
                    dv_sens_pmf_renorm=self.dv_sens_pmf_renorm,
                    dv_meta_mode=self.dv_meta_mode,
                    dv_sens_mode=self.dv_sens_mode,
                    confidence_mode=confidence_mode,  # noqa
                    choiceprob=self.choiceprob,
                    posterior=self.posterior,
                )
                model_extended = make_dataclass('ModelExtended', dict_extended.keys())
                model_extended.__module__ = '__main__'
                desc.update(dict(extended=model_extended(**dict_extended)))

        model_summary = make_dataclass('ModelSummary', desc.keys())

        def _repr(self_):
            txt = f'{self_.__class__.__name__}'
            for k, v in self_.__dict__.items():
                if k in ('data', 'fit'):
                    txt += f"\n\t{k}: <not displayed>"
                elif k == 'extended':
                    txt += f"\n\t{k}: additional modeling results (attributes: " \
                           f"{', '.join([a for a in self_.extended.__dict__.keys()])})"
                else:
                    txt += f"\n\t{k}: {v}"
            return txt

        model_summary.__repr__ = _repr
        model_summary.__module__ = '__main__'
        return model_summary(**desc)


class ModelFit(ReprMixin):
    fit_sens: OptimizeResult = None
    fit_meta: OptimizeResult = None



import warnings
from dataclasses import dataclass
from typing import Callable, Dict, Union, List

import numpy as np

from .modelspec import Parameter, ParameterSet
from .util import ReprMixin, _slsqp_epsilon
from importlib.util import find_spec


@dataclass
class Configuration(ReprMixin):
    """
    Configuration for the ReMeta toolbox

    Parameters
    ----------
    *** Basic definition of the model ***
    skip_meta : bool
        If False, only fit the sensory level.
    meta_noise_type : str
        Whether the model considers noise at readout or report.
        Possible values: 'noisy_report', 'noisy_readout'
    meta_noise_dist : str
        Metacognitive noise distribution.
        Possible valus: 'truncated_norm', 'truncated_gumbel', 'truncated_lognorm',
                        'norm', 'gumbel', 'lognorm', 'beta', 'gamma',
                        'censored_norm', 'censored_gumbel', 'censored_lognorm', 'censored_gamma'
                        Experimental (for internal use only):
                        'beta_std', 'betaprime', 'lognorm_varstd',
                        'censored_lognorm', 'censored_lognorm_varstd', 'censored_betaprime',
                        'truncated_norm_fit', 'truncated_lognorm_varstd',
                        Works only with lookup table (not deployed via pip):
                        'truncated_norm_lookup', 'truncated_gumbel_lookup'
    meta_link_function : str
        Metacognitive link function. In case of criterion-based link functions {x} refers to the number of criteria.
        Possible values: 'probability_correct', 'identity', 'tanh', 'normcdf', 'erf', 'alg', 'guder', 'linear',
                         '{x}_criteria', '{x}_criteria_linear', '{x}_criteria_linear_tanh'
                         Note: replace {x} with the number of criteria
                         Experimental (for internal use only):
                         'detection_model_linear', 'detection_model_mean', 'detection_model_mode',
                         'detection_model_full', 'detection_model_ideal'

    detection_model : bool
        Experimental. Define the model as a detection model.
    detection_model_nchannels : int
        Defines the number of sensory channels in case of a detection model.

    *** Enable or disable specific parameters ***
    * Each setting can take the values 0, 1 or 2:
    *    0: Disable parameter.
    *    1: Enable parameter.
    *    2: Enable parameter and fit separate values for the negative and positive stimulus category (in the case of
    *       sensory parameters, indicated by the suffix '_sens') or negative and positive decision values (in the
    *       case of metacognitive parameters, indicated by the suffix '_meta').
    enable_warping_sens : int (default: 0)
        Fit a non-linear transducer prior to sensory processing (the nonlinear function is defined via
        `function_warping_sens`).
    enable_noise_sens : int (default: 1)
        Fit separate sensory noise parameters for both stimulus categories.
    enable_noise_transform_sens : int (default: 0)
        Fit an additional sensory noise parameter for signal-dependent sensory noise (the type of dependency is is
        defined via `function_noise_transform_sens`).
    enable_thresh_sens : int (default: 0)
        Fit a sensory threshold.
    enable_bias_sens : int (default: 1)
        Fit a sensory bias towards one of the stimulus categories.
    enable_noise_meta : int (default: 1)
        Fit a metacognitive noise parameter
    enable_noise_transform_meta : int (default: 0)
        Fit an additional metacognitive noise parameter for signal-dependent metacognitive noise (the type of dependency
        is defined via `function_noise_transform_meta`). Note: at present, enable_noise_transform_meta=2 leads to
        biased results and is therefore discouraged.
    enable_evidence_bias_add_meta : int (default: 0)
        Fit an additive metacognitive bias loading on sensory evidence.
    evidence_bias_mult_meta : int (default: 1)
        Fit a multiplicative metacognitive bias loading on evidence. In the case of a criterion-based link function,
        multiple parameters are fitted for each criterion/confidence level.
    evidence_bias_mult_postnoise_meta : int (default: 0)
        Fit a multiplicative metacognitive bias loading on evidence, but operating after the application of readout
        noise. In the case of a criterion-based link function, multiple parameters are fitted for each
        criterion/confidence level.
    enable_confidence_bias_mult_meta : int (default: 0)
        Fit a multiplicative metacognitive bias loading on confidence.
    enable_confidence_bias_add_meta : int (default: 0)
        Fit an additive metacognitive bias loading on confidence.
    enable_confidence_bias_pow_meta : int (default: 1)
        Fit an exponential metacognitive bias loading on confidence.
    enable_criteria_meta : int (default: 0)
        Fit criteria for a criterion-based link function. Note that the number of criteria is set via the link function
        argument (see `meta_link_function`)
    enable_levels_meta : int (default: 0)
        Fit confidence levels for a criterion-based link function. If disabled, equidistant confidence levels are
        assumed (e.g. 1/3, 2/3, 1 for three criteria).


    *** Define fitting characteristics of the parameters ***
    * The fitting of each parameter is characzerized as follows:
    *     1) An initial guess.
    *     2) Lower and upper bound.
    *     3) Grid range, i.e. the range of values that are tested during the initial gridsearch search.
    * Sensible default values are provided for all parameters. To tweak those, one can either define an entire
    * ParameterSet, which is a container for a set of parameters, or each parameter individually. Note that the
    * parameters must be either defined as a Parameter instance or as List[Parameter] in case when separate values are
    * fitted for the positive and negative stimulus category/decision value).
    paramset_sens : ParameterSet
        Parameter set for the sensory level.
    paramset_meta : ParameterSet
        Parameter set for the metacognitive level.

    warping_sens : Union[Parameter, List[Parameter]]
        Parameter for the nonlinear transducer.
    noise_sens : Union[Parameter, List[Parameter]]
        Parameter for sensory noise.
    noise_transform_sens : Union[Parameter, List[Parameter]]
        Parameter for multiplicative sensory noise.
    thresh_sens : Union[Parameter, List[Parameter]]
        Parameter for the sensory threshold.
    bias_sens : Union[Parameter, List[Parameter]]
        Parameter for the sensory bias.
    noise_meta : Union[Parameter, List[Parameter]]
        Parameter for metacognitive noise.
    noise_transform_meta : Union[Parameter, List[Parameter]]
        Parameter for multiplicative megacognitive noise.
    evidence_bias_mult_meta : Union[Parameter, List[Parameter]]
        Parameter for multiplicative metacognitive bias loading on evidence.
    evidence_bias_add_meta : Union[Parameter, List[Parameter]]
        Parameter for additive metacognitive bias loading on evidence.
    evidence_bias_mult_postnoise_meta : Union[Parameter, List[Parameter]]
        Parameter for multiplicative metacognitive bias loading on evidence, but after the application of readout noise.
    confidence_bias_mult_meta : Union[Parameter, List[Parameter]]
        Parameter for multiplicative metacognitive bias loading on confidence.
    confidence_bias_add_meta : Union[Parameter, List[Parameter]]
        Parameter for additive metacognitive bias loading on confidence.
    confidence_bias_pow_meta : Union[Parameter, List[Parameter]]
        Parameter for an exponential metacognitive bias loading on confidence.
    criterion{x}_meta : Union[Parameter, List[Parameter]]
        Parameter for the xth confidence criterion (0-based).
    level{x}_meta : Union[Parameter, List[Parameter]]
        Parameter for the xth confidence level (0-based).

    *** Constraints ***
    * Constraints for the parameters are provided in the form of a function, as defined below.
    constraints_sens_callable : Callable
        Constraints for sensory parameters. See method `_constraints_sens_default` in this module for the expected
        format of this Callable. See method `_constraints_sens_default` in this module for the expected
        format of this Callable. The function must return a list of dictionaries, where each
        dictionary specifies a scipy optimize constraint.
    constraints_meta_callable : Callable
        Constraints for metacognitive parameters. The function must take the configuration object as the first argument
        and the stimulus array as the second argument. The function must return a list of dictionaries, where each
        dictionary specifies a scipy optimize constraint.

    *** Methodoligcal aspects of parameter fitting ***
    * Note: this applies to the fitting of metacognitive parameters only.
    gridsearch : bool (default: True)
        If True, perform initial (usually coarse) gridsearch search, based on the gridsearch defined for a Parameter.
    fine_gridsearch : bool (default: False)
        If True, perform an iteratively finer gridsearch search for each parameter.
    grid_multiproc : bool (default: False)
        If True, use all available cores for the gridsearch search. If False, use a single core.
    global_minimization : bool (default: False)
        If True, use a global minimization routine.
    gradient_method : str or Tuple/List (default: 'slsqp')
        Set scipy.optimize.minimize gradient method
        If provided as Tuple/List, test different gradient methods and take the best
    gradient_free : bool (default: None)
        If True, use a gradien-free optimization routine.
    slsqp_epsilon : float or Tuple/List (default: None)
        Set parameter epsilon parameter for the SLSQP optimization method.
        If provided as Tuple/List, test different eps parameters and take the best
    init_nelder_mead : bool (default: False)
        If True, jump-start parameter minimization with the gradient-free Nelder-Mead algorithm

    *** Transformation functions ***
    function_warping_sens: str (default: 'power')
        Can be one of 'power', 'exponential' or 'identity'.
    function_noise_transform_sens: str (default: 'multiplicative')
        Can be one of 'multiplicative', 'power', 'exponential' or 'logarithm'.
    function_noise_transform_meta: str (default: 'multiplicative')
        Can be one of 'multiplicative', 'power', 'exponential' or 'logarithm'.

    *** Preprocessing ***
    normalize_stimuli_by_max : bool (default: True)
        If True, normalize provided stimuli by their maximum value.
    confidence_bounds_error : float
        Set confidence < confidence_bounds_error to 0 and > confidence_bounds_erroror to 1. This might be useful
        for continuous scales, if 'finger errors' are assumed responsible for the fact that confidence sometimes is not
        *exactly* 0 or 1 but very close (i.e. the assumption is that observers intended to indicate a confidence of 0/1
        but failed to do so exactly)

    *** Parameters for the metacognitive likelihood computation ***
    min_likelihood_sens : float
        Minimum probability used during the sensory likelihood computation
    min_likelihood_meta : float
        Minimum probability used during the metacognitive likelihood computation
    binsize_meta : float
        Integration bin size for the computation of the likelihood around empirical confidence values (noisy-report)
        or metacognitive evidence (noisy-readout).
    max_dv_deviation : int
        Number of standard deviations around the mean considered for sensory uncertainty.
    nbins_dv : int
        Number of discrete decision values bins that are considered to represent sensory uncertainty.
    experimental_likelihood : bool
        Instead of using a minimum probability during the likelihood computation, use a maximum cumulative
        likelihood based on a 'guessing' model
    experimental_wrap_binsize_meta : bool (default: False)
        Ensure constant window size for likelihood integration at the bounds
    experimental_include_incongruent_dv : bool (default: False)
        Include incongruent decision values (i.e., sign(actual choice) != sign(decision value)) for the likelihood
        computation

    *** Other ***
    true_params : Dict
        Pass true (known) parameter values. This can be useful for testing to compare the likelihood of true and
        fitted parameters. The likelihood of true parameters is returned (and printed).
    force_settings : bool
        Some setting combinations are known to be incompatible and/or to produce biased fits. If True, fit the model
        nevertheless.
    settings_ignore_warnings : bool (default: False)
        If True, ignore warnings about user-specified settings.
    print_configuration : bool (default: True)
        If True, print the configuration at instatiation of the ReMeta class.
    """

    skip_meta: bool = False
    meta_noise_type: str = 'noisy_report'
    meta_noise_dist: str = 'truncated_norm'
    meta_link_function: str = 'probability_correct'
    detection_model: bool = False
    detection_model_nchannels: int = 10

    enable_warping_sens: int = 0
    enable_noise_sens: int = 1
    enable_thresh_sens: int = 0
    enable_bias_sens: int = 1
    enable_noise_meta: int = 1
    enable_evidence_bias_mult_meta: int = 1
    enable_evidence_bias_add_meta: int = 0
    enable_confidence_bias_mult_meta: int = 0
    enable_confidence_bias_add_meta: int = 0
    enable_confidence_bias_pow_meta: int = 0
    enable_criteria_meta: int = 0
    enable_levels_meta: int = 0
    # Experimental:
    enable_noise_transform_sens: int = 0
    enable_noise_transform_meta: int = 0
    enable_evidence_bias_mult_postnoise_meta: int = 0

    paramset_sens: ParameterSet = None
    paramset_meta: ParameterSet = None

    warping_sens: Union[Parameter, List[Parameter]] = None
    noise_sens: Union[Parameter, List[Parameter]] = None
    noise_transform_sens: Union[Parameter, List[Parameter]] = None
    thresh_sens: Union[Parameter, List[Parameter]] = None
    bias_sens: Union[Parameter, List[Parameter]] = None
    noise_meta: Union[Parameter, List[Parameter]] = None
    noise_transform_meta: Union[Parameter, List[Parameter]] = None
    evidence_bias_mult_meta: Union[Parameter, List[Parameter]] = None
    evidence_bias_add_meta: Union[Parameter, List[Parameter]] = None
    evidence_bias_mult_postnoise_meta: Union[Parameter, List[Parameter]] = None
    confidence_bias_mult_meta: Union[Parameter, List[Parameter]] = None
    confidence_bias_add_meta: Union[Parameter, List[Parameter]] = None
    confidence_bias_pow_meta: Union[Parameter, List[Parameter]] = None

    constraints_sens_callable: Callable = None
    constraints_meta_callable: Callable = None

    function_warping_sens: str = 'power'
    function_noise_transform_sens: str = 'multiplicative'
    function_noise_transform_meta: str = 'multiplicative'

    gridsearch: bool = True
    fine_gridsearch: bool = False
    grid_multiproc: bool = False
    global_minimization: bool = False
    gradient_method: str = 'slsqp'
    gradient_free: bool = None
    slsqp_epsilon: float = None
    init_nelder_mead: bool = False

    normalize_stimuli_by_max: bool = True
    confidence_bounds_error: float = 0

    min_likelihood_meta: float = 1e-10
    min_likelihood_sens: float = 1e-10
    binsize_meta: float = 1e-1
    max_dv_deviation: int = 5
    nbins_dv: int = 101
    experimental_likelihood: bool = False
    experimental_wrap_binsize_meta: bool = False
    experimental_include_incongruent_dv: bool = False

    true_params: Dict = None
    force_settings: bool = False
    settings_ignore_warnings: bool = False
    print_configuration: bool = False

    noise_sens_default: float = 0.001
    noise_meta_default: float = 0.1
    noise_meta_min: float = 0.001

    _warping_sens_default: Parameter = Parameter(guess=0.1, bounds=(-10, 10), grid_range=np.arange(-10, 11, 5))
    _noise_transform_sens_default: Parameter = Parameter(guess=0, bounds=(0, 10), grid_range=np.arange(0, 1.1, 0.25))
    _noise_sens_default: Parameter = Parameter(guess=0.1, bounds=(1e-3, 100), grid_range=np.arange(0.1, 0.9, 0.25))
    _thresh_sens_default: Parameter = Parameter(guess=0, bounds=(0, 1), grid_range=np.arange(0, 0.41, 0.2))
    _bias_sens_default: Parameter = Parameter(guess=0, bounds=(-1, 1), grid_range=np.arange(-0.1, 0.11, 0.1))
    _noise_meta_default: Parameter = Parameter(guess=0.2, bounds=(1e-2, 1), grid_range=np.arange(0.05           , 1, 0.1))
    _noise_transform_meta_default: Parameter = Parameter(guess=0, bounds=(0, 10),
                                                         grid_range=np.arange(0, 1.1, 0.25))
    _evidence_bias_mult_meta_default: Parameter = Parameter(guess=1, bounds=(0.5, 2),
                                                            grid_range=np.arange(0.5, 2.01, 0.1))
    _evidence_bias_add_meta_default: Parameter = Parameter(guess=0, bounds=(-0.5, 0.5),
                                                           grid_range=np.arange(-0.2, 0.21, 0.1))
    _evidence_bias_mult_postnoise_meta_default: Parameter = Parameter(guess=1, bounds=(0.1, 50),
                                                                      grid_range=np.arange(0.2, 1.71, 0.3))
    _confidence_bias_mult_meta_default: Parameter = Parameter(guess=1, bounds=(0.1, 10),
                                                              grid_range=np.arange(0.5, 2.01, 0.3))
    _confidence_bias_add_meta_default: Parameter = Parameter(guess=0, bounds=(-1, 1),
                                                             grid_range=np.arange(-0.3, 0.31, 0.15))
    _confidence_bias_pow_meta_default: Parameter = Parameter(guess=1, bounds=(0.2, 5),
                                                             grid_range=np.arange(0.5, 2.1, 0.25))
    _criterion_meta_default: Parameter = Parameter(guess=0, bounds=(1e-6, 50),
                                                   grid_range=np.exp(np.linspace(0, np.log(2), 8)) - 0.9)
    _level_meta_default: Parameter = Parameter(guess=0, bounds=(1e-6, 1),
                                               grid_range=np.exp(np.linspace(0, np.log(2), 8)) - 0.9)

    def setup(self):

        if self.gradient_free is None:
            if '_criteria' in self.meta_link_function or '_transform' in self.meta_noise_dist:
                self.gradient_free = True
            else:
                self.gradient_free = False

        if self.slsqp_epsilon is None:
            if self.meta_noise_type == 'noisy_readout':
                self.slsqp_epsilon = 1e-4
            else:
                self.slsqp_epsilon = _slsqp_epsilon

        if self._confidence_bias_mult_meta_default is None and (self.meta_noise_type == 'noisy_readout'):
            self._confidence_bias_mult_meta_default = Parameter(guess=1, bounds=(0.1, 10),
                                                                grid_range=np.arange(0.4, 1.01, 0.2))

        if find_spec('multiprocessing_on_dill') is None:
            warnings.warn(f'Multiprocessing on dill is not installed. Setting grid_multiproc is changed to False.')
            self.grid_multiproc = False

        self._check_compatibility()

        self._prepare_params_sens()
        self._prepare_params_meta()
        if self.print_configuration:
            self.print()

    def _check_compatibility(self):

        if self.enable_noise_transform_meta:
            text = 'Fitting signal-dependent metacognitive noise parameters leads to biased estimates (for currently ' \
                   'unknown reasons) and is thus discouraged.'
            if self.force_settings:
                warnings.warn(text)
            else:
                raise ValueError(text)
        # if self.enable_noise_transform_meta == 2:
        #     warnings.warn('Fitting separate signal-dependent metacognitive noise parameters for the two stimulus '
        #                   'categories leads to biased estimates (for currently unknown reasons) and is thus '
        #                   'discouraged.')

        if self.detection_model and 'detection_model' not in self.meta_link_function:
            raise ValueError('Detection Models require a compatible metacogntive link function. Choose one of: '
                             'detection_model_linear, detection_model_mean, detection_model_mode, '
                             'detection_model_full, detection_model_ideal')

        if not self.enable_criteria_meta and '_criteria' in self.meta_link_function:
            raise ValueError('A criterion-based link function was set, but confidence criteria were not enabled.')

        if not self.settings_ignore_warnings:

            if not self.enable_noise_meta:
                warnings.warn(f'Setting enable_noise_meta=False was provided -> noise_meta is set to its default value '
                              f'({self.noise_meta_default}). You may change this value via the configuration.')

            if self.enable_criteria_meta and '_criteria' not in self.meta_link_function:
                self.enable_criteria_meta = 0
                warnings.warn('Confidence criteria were enabled but no criterion-based link function was set -> '
                              'auto-setting enable_criteria_meta = 0.')
            if self.enable_levels_meta and '_criteria' not in self.meta_link_function:
                self.enable_levels_meta = 0
                warnings.warn('Confidence criteria were enabled but no criterion-based link function was set -> '
                              'auto-setting enable_levels_meta = 0.')
            if self.enable_levels_meta and not self.enable_criteria_meta:
                self.enable_criteria_meta = self.enable_levels_meta
                warnings.warn(f'Confidence leves were enabled, but confidence criteria not -> auto-setting'
                              f'enable_criteria_meta = {self.enable_levels_meta}')

            if self.meta_noise_type == 'noisy_readout':
                if self.enable_confidence_bias_add_meta:
                    warnings.warn('The setting enable_confidence_bias_add_meta has been enabled for a model of type '
                                  'noisy-readout. This setting results in a link function that is not monotonically '
                                  'increasing and likely leads to biased estimates.')
                if self.enable_confidence_bias_mult_meta:
                    warnings.warn('The setting enable_confidence_bias_mult_meta has been enabled for a model of type '
                                  'noisy-readout. Use this only if you have strong reasons to belief that values of '
                                  'this parameter are <= 1, since true values > 1 cannot be recovered for '
                                  'noisy-readout models.')

            if self.enable_criteria_meta and self.enable_evidence_bias_mult_meta:
                self.enable_evidence_bias_mult_meta = 0
                warnings.warn('Confidence criteria were enabled, which is in conflict with '
                              'enable_evidence_bias_mult_meta > 0 -> auto-setting enable_evidence_bias_mult_meta = 0.')

            if self.enable_criteria_meta and self.enable_evidence_bias_mult_postnoise_meta:
                self.enable_evidence_bias_mult_postnoise_meta = 0
                warnings.warn('Confidence criteria were enabled, which is in conflict with '
                              'enable_evidence_bias_mult_postnoise_meta > 0 -> auto-setting '
                              'enable_evidence_bias_mult_postnoise_meta = 0.')

    def _prepare_params_sens(self):
        if self.paramset_sens is None:

            param_names_sens = []
            params_sens = ('warping', 'noise', 'noise_transform', 'thresh', 'bias')
            for param in params_sens:
                if getattr(self, f'enable_{param}_sens'):
                    param_names_sens += [f'{param}_sens']
                    if getattr(self, f'{param}_sens') is None:
                        param_default = getattr(self, f'_{param}_sens_default')
                        if getattr(self, f'enable_{param}_sens') == 2:
                            setattr(self, f'{param}_sens', [param_default, param_default])
                        else:
                            setattr(self, f'{param}_sens', param_default)

            parameters = {k: getattr(self, k) for k in param_names_sens}
            self.paramset_sens = ParameterSet(parameters, param_names_sens)

        if self.constraints_sens_callable is None:
            self.constraints_sens_callable = _constraints_sens_default

    def _prepare_params_meta(self):

        if self.paramset_meta is None:

            if self.enable_noise_meta and self.noise_meta is None:
                if self.meta_noise_dist == 'beta':
                    self._noise_meta_default.bounds = (1e-5, 0.5)
                    self._noise_meta_default.grid_range = np.arange(0.05, 0.5, 0.05)
                elif self.meta_noise_type == 'noisy_readout':
                    self._noise_meta_default.bounds = (1e-5, 250)

            param_names_meta = []
            params_meta = ('noise', 'noise_transform', 'evidence_bias_mult', 'evidence_bias_add',
                           'evidence_bias_mult_postnoise', 'confidence_bias_mult', 'confidence_bias_add',
                           'confidence_bias_pow')
            for param in params_meta:
                if getattr(self, f'enable_{param}_meta'):
                    param_names_meta += [f'{param}_meta']
                    if getattr(self, f'{param}_meta') is None:
                        param_default = getattr(self, f'_{param}_meta_default')
                        if getattr(self, f'enable_{param}_meta') == 2:
                            setattr(self, f'{param}_meta', [param_default.copy(), param_default.copy()])
                        else:
                            setattr(self, f'{param}_meta', param_default.copy())

            if self.enable_criteria_meta and '_criteria' in self.meta_link_function:  # noqa
                ncriteria_meta = int(self.meta_link_function.split('_')[0])
                guess_criteria = np.linspace(1 / (ncriteria_meta + 1), 1 - 1 / (ncriteria_meta + 1), ncriteria_meta)

                for i in range(ncriteria_meta):
                    if not hasattr(self, f'criterion{i}_meta') or getattr(self, f'criterion{i}_meta') is None:  # noqa
                        param_names_meta += [f'criterion{i}_meta']
                        param_default = self._criterion_meta_default.copy()
                        param_default.guess = guess_criteria[i]
                        if self.enable_criteria_meta == 2:
                            setattr(self, f'criterion{i}_meta', [param_default.copy(), param_default.copy()])
                        else:
                            setattr(self, f'criterion{i}_meta', param_default)
                    if self.enable_levels_meta and \
                            (not hasattr(self, f'level{i}_meta') or getattr(self, f'level{i}_meta') is None):
                        param_names_meta += [f'level{i}_meta']
                        param_default = self._level_meta_default.copy()
                        param_default.guess = guess_criteria[i]
                        if self.enable_levels_meta == 2:
                            setattr(self, f'level{i}_meta', [param_default.copy(), param_default.copy()])
                        else:
                            setattr(self, f'level{i}_meta', param_default)

            parameters = {k: getattr(self, k) for k in param_names_meta}
            self.paramset_meta = ParameterSet(parameters, param_names_meta)

        if self.constraints_meta_callable is None:
            self.constraints_meta_callable = _constraints_meta_default

    def print(self):
        # print('***********************')
        print(f'{self.__class__.__name__}')
        for k, v in self.__dict__.items():
            if not self.skip_meta or ('meta_' not in k and '_meta' not in k):
                print('\n'.join([f'\t{k}: {v}']))
        # print('***********************')

    def __repr__(self):
        txt = f'{self.__class__.__name__}\n'
        txt += '\n'.join([f'\t{k}: {v}' for k, v in self.__dict__.items()])
        return txt


def _constraints_sens_default(remeta):
    constraints = []
    if remeta.cfg.enable_noise_transform_sens:
        constraints += [{'type': 'ineq', 'fun': lambda x: np.min(remeta._negll_sens(x, return_noise=True)) - 0.001}]  # noqa
    return constraints


def _constraints_meta_default(remeta):
    constraints = []

    # Set constraints for the metacognitive noise slope
    # Note that metacognitive noise is based on confidence in case of the noisy-report model and
    # on metacognitive decision values in case of the noisy-readout model
    if remeta.cfg.enable_noise_transform_meta:
        constraints += [{'type': 'ineq', 'fun': lambda x: np.min(remeta.fun_meta_helper(x, return_noise=True,
                                                                                        constraint_mode=True)) - 0.001}]

    if '_criteria' in remeta.cfg.meta_link_function:
        ncrit_meta = int(remeta.cfg.meta_link_function.split('_')[0])
        # in case of a linear_tanh link function, the last parameter criterion corresponds to the slope of the tanh
        ncrit_meta_ = ncrit_meta - 1 if 'linear_tanh' in remeta.cfg.meta_link_function else ncrit_meta
        if remeta.cfg.enable_criteria_meta == 2:
            def fun_criteria(x):
                crit = remeta.fun_meta_helper(x, return_criteria=True, constraint_mode=True)
                return np.sum([(crit[0][i] > (0 if i == 0 else crit[0][i - 1])) for i in range(ncrit_meta_)]) + \
                    np.sum([(crit[1][i] > (0 if i == 0 else crit[1][i - 1])) for i in range(ncrit_meta_)]) - \
                    2 * ncrit_meta_
        else:
            def fun_criteria(x):
                crit = remeta.fun_meta_helper(x, return_criteria=True, constraint_mode=True)
                return np.sum([(crit[i] > (0 if i == 0 else crit[i - 1])) for i in range(ncrit_meta_)]) - ncrit_meta_
        constraints += [{'type': 'eq', 'fun': fun_criteria}]
        if remeta.cfg.enable_levels_meta:
            if remeta.cfg.enable_levels_meta == 2:
                def fun_levels(x):
                    lev = remeta.fun_meta_helper(x, return_levels=True, constraint_mode=True)
                    return np.sum([(lev[0][i] > (0 if i == 0 else lev[0][i - 1])) for i in range(ncrit_meta)]) + \
                        np.sum([(lev[1][i] > (0 if i == 0 else lev[1][i - 1])) for i in range(ncrit_meta)]) - \
                        2 * ncrit_meta
            else:
                def fun_levels(x):
                    lev = remeta.fun_meta_helper(x, return_levels=True, constraint_mode=True)
                    return np.sum([(lev[i] > (0 if i == 0 else lev[i - 1])) for i in range(ncrit_meta)]) - ncrit_meta
            constraints += [{'type': 'eq', 'fun': fun_levels}]
    return constraints



import timeit
import warnings
from functools import partial
from itertools import product
from multiprocessing import cpu_count

import numpy as np
try:  # only necessary if multiple cores should be used
    from multiprocessing_on_dill.pool import Pool as DillPool
except ModuleNotFoundError:
    pass
from scipy.optimize import OptimizeResult
from scipy.optimize import minimize, Bounds, basinhopping
from scipy.optimize._constraints import old_bound_to_new  # noqa
from .util import TAB
from .util import _slsqp_epsilon


def loop(fun, params, args, gridsize, verbose, param_id):
    # t0 = timeit.default_timer()
    if param_id and verbose and (np.mod(param_id, 1000) == 0):
        print(f'{TAB}{TAB}Grid iteration {param_id} / {gridsize}')
    result = fun(params[param_id], *args)
    # print(timeit.default_timer() - t0)
    return result


def fgrid(fun, valid, args, grid_multiproc, verbose=True):
    if grid_multiproc:
        with DillPool(cpu_count() - 1 or 1) as pool:
            ll_grid = pool.map(partial(loop, fun, valid, args, len(valid), verbose), range(len(valid)))
    else:
        ll_grid = [None] * len(valid)
        for i, param in enumerate(valid):
            ll_grid[i] = loop(fun, valid, args, len(valid), verbose, i)
    return ll_grid


def grid_search(x0, fun, args, param_set, ll_grid, valid, n_grid_candidates, n_grid_iter, grid_multiproc, verbose=True):
    n_grid_candidates_init = 3 * n_grid_candidates
    gx0 = x0
    gll_min_grid = np.min(ll_grid)
    grid_candidates = [valid[i] for i in np.argsort(ll_grid)[:n_grid_candidates_init]]
    previous_grid_range = [param_set.grid_range for _ in range(n_grid_candidates_init)]
    candidate_ids = list(range(n_grid_candidates_init))
    counter = 0
    for i in range(n_grid_iter):
        if verbose:
            print(f'{TAB}Grid iteration {i + 1} / {n_grid_iter}')
        gvalid = []
        valid_candidate_ids = []
        grid_range = [None] * len(grid_candidates) * 2
        for j, grid_candidate in enumerate(grid_candidates):
            grid_range[j] = [None] * param_set.nparams
            for k, p in enumerate(grid_candidate):
                ind = np.where(previous_grid_range[candidate_ids[j]][k] == p)[0][0]
                lb = previous_grid_range[candidate_ids[j]][k][max(0, ind - 1)]  # noqa
                ub = previous_grid_range[candidate_ids[j]][k][  # noqa
                    min(len(previous_grid_range[candidate_ids[j]][k]) - 1, ind + 1)]  # noqa
                grid_range[j][k] = np.around(np.linspace(lb, ub, 5), decimals=10)  # noqa
            gvalid_candidate = [p for p in product(*grid_range[j]) if  # noqa
                                np.all([con['fun'](p) >= 0 for con in param_set.constraints])]
            gvalid += gvalid_candidate
            valid_candidate_ids += [j] * len(gvalid_candidate)
            if verbose:
                print(f'{TAB}{TAB}Candidate {j + 1}: {[(p[0], p[-1]) for p in grid_range[j]]}')  # noqa
        gll_grid = fgrid(fun, gvalid, args, grid_multiproc, verbose=False)
        counter += len(gvalid)

        min_id = np.argmin(gll_grid)
        if gll_grid[min_id] < gll_min_grid:
            gll_min_grid = gll_grid[min_id]
            gx0 = gvalid[min_id]

        if verbose:
            print(f'{TAB}{TAB}Best fit: {gx0} (LL={gll_min_grid})')

        if i != n_grid_iter - 1:
            grid_candidates, candidate_ids = [], []
            for j in np.argsort(gll_grid):
                if gvalid[j] not in grid_candidates:
                    grid_candidates += [gvalid[j]]
                    candidate_ids += [valid_candidate_ids[j]]
                    if len(candidate_ids) == n_grid_candidates:
                        break
            previous_grid_range = grid_range
    fit = OptimizeResult(success=True, x=gx0, fun=gll_min_grid, nfev=counter)

    return fit


def fmincon(fun, param_set, args, gridsearch=False, grid_multiproc=True,
            gradient_free=False, global_minimization=False, fine_gridsearch=False, verbose=True,
            n_grid_candidates=10, n_grid_iter=3, gradient_method='slsqp', slsqp_epsilon=_slsqp_epsilon,
            init_nelder_mead=False, guess=None):

    t0 = timeit.default_timer()

    if verbose:
        negll_initial_guess = fun(param_set.guess, *args)
        print(f'Initial guess (neg. LL: {negll_initial_guess:.2f})')
        for i, p in enumerate(param_set.names):
            print(f'{TAB}[guess] {p}: {param_set.guess[i]:.4g}')

    bounds = Bounds(*old_bound_to_new(param_set.bounds), keep_feasible=True)
    if gridsearch:
        if len(param_set.constraints):
            valid = [p for p in product(*param_set.grid_range) if
                     np.all([con['fun'](p) >= 0 for con in param_set.constraints])]
        else:
            valid = list(product(*param_set.grid_range))
        if verbose:
            print(f"Grid search activated (grid size = {len(valid)})")
        t0 = timeit.default_timer()
        ll_grid = fgrid(fun, valid, args, grid_multiproc, verbose=verbose)
        x0 = valid[np.argmin(ll_grid)]
        ll_min_grid = np.min(ll_grid)
        grid_time = timeit.default_timer() - t0
        if verbose:
            for i, p in enumerate(param_set.names):
                print(f'{TAB}[grid] {p}: {x0[i]:.4g}')
            print(f"Grid neg. LL: {ll_min_grid:.1f}")
            print(f"Grid runtime: {grid_time:.2f} secs")
        fit_grid = OptimizeResult(success=True, x=x0, fun=ll_min_grid, nfev=len(valid))
    else:
        if guess is None:
            x0 = param_set.guess
        else:
            x0 = guess
        fit_grid = OptimizeResult(success=True, x=x0, fun=fun(x0, *args), nfev=1)

    if fine_gridsearch:
        x0 = grid_search(x0, fun, args, param_set, ll_grid, valid, n_grid_candidates, n_grid_iter,  # noqa
                         grid_multiproc, verbose=verbose).x

    if gradient_free:
        if global_minimization:
            if verbose:
                print('Performing global optimization')
            fit = basinhopping(fun, x0,
                               take_step=RandomDisplacementBoundsConstraints(bounds, param_set.constraints),
                               accept_test=BoundsConstraints(bounds, param_set.constraints),
                               minimizer_kwargs=dict(method='Nelder-Mead', args=tuple(args)))
        else:
            if verbose:
                print('Performing local optimization')

            fit_nm = minimize(fun, fit_grid.x, args=tuple(args), method='Nelder-Mead')
            if verbose:
                print(f'LL Nelder_mead: {fit_nm.fun:.3f}')
            use_grid_guess = False
            if fit_grid.fun < fit_nm.fun:
                use_grid_guess = True
                warnings.warn(f'Grid guess (LL={fit_grid.fun:.3f}) superior to minimize (LL={fit_nm.fun:.3f})')
            bds = [(fit_nm.x[i] >= b[0]) & (fit_nm.x[i] <= b[1]) for i, b in enumerate(param_set.bounds)]
            cns = [con['fun'](fit_nm.x) >= 0 for con in param_set.constraints]
            if not (np.all(cns) and np.all(bds)):
                use_grid_guess = True
                warnings.warn(f'Nelder-Mead estimate {fit_nm.x} violates constraints {not np.all(cns)} and/or '
                              f'bounds {not np.all(bds)}')
            if use_grid_guess:
                fit = fit_grid
            else:
                fit = fit_nm
    else:
        if global_minimization:
            if verbose:
                print('Performing global optimization')
            fit = basinhopping(fun, x0, take_step=RandomDisplacementBoundsConstraints(bounds, param_set.constraints),
                               accept_test=BoundsConstraints(bounds, param_set.constraints),
                               minimizer_kwargs=dict(method='Nelder-Mead', args=tuple(args)))
        else:
            if verbose:
                print('Performing local optimization')

            if init_nelder_mead:
                fit_nelder = minimize(fun, x0, bounds=bounds, args=tuple(args), constraints=param_set.constraints,
                                      method='nelder-mead')
                for i, b in enumerate(param_set.bounds):
                    fit_nelder.x[i] = min(max(fit_nelder.x[i], b[0]), b[1])
                if fun(fit_nelder.x, *args) < fun(x0, *args):
                    x0 = fit_nelder.x
            gradient_method = [gradient_method] if isinstance(gradient_method, str) else gradient_method
            min_fun = fun(x0, *args)
            fit = fit_grid
            for method in gradient_method:
                if method == 'slsqp':
                    slsqp_epsilon_ = slsqp_epsilon if hasattr(slsqp_epsilon, '__len__') else [slsqp_epsilon]
                    for eps in slsqp_epsilon_:
                        fit_ = minimize(fun, x0, bounds=bounds, args=tuple(args), constraints=param_set.constraints,
                                        method='slsqp', options=dict(eps=eps))
                        if fit_.fun < min_fun:
                            min_fun = fit_.fun
                            fit = fit_
                            # x0 = fit_.x  # risk of local minima?
                else:
                    fit_ = minimize(fun, x0, bounds=bounds, args=tuple(args), constraints=param_set.constraints,
                                    method=method)
                    if fit_.fun < min_fun:
                        min_fun = fit_.fun
                        fit = fit_
                        # x0 = fit_.x  # risk of local minima?
    fit.execution_time = timeit.default_timer() - t0  # noqa

    return fit


class RandomDisplacementBoundsConstraints(object):
    def __init__(self, bounds, constraints, stepsize=0.5):
        self.xmin = bounds.lb
        self.xmax = bounds.ub
        self.constraints = constraints
        self.stepsize = stepsize

    def __call__(self, x):
        while True:
            xnew = x + np.random.uniform(-self.stepsize, self.stepsize, np.shape(x))
            if np.all(xnew < self.xmax) and np.all(xnew > self.xmin) and np.all(
                    [con['fun'](xnew) >= 0 for con in self.constraints]):
                break
        return xnew


class BoundsConstraints(object):

    def __init__(self, bounds, constraints):
        self.xmin = bounds.lb
        self.xmax = bounds.ub
        self.constraints = constraints

    def __call__(self, **kwargs):
        x = kwargs["x_new"]
        tmax = bool(np.all(x <= self.xmax))
        tmin = bool(np.all(x >= self.xmin))
        cons = bool(np.all([con['fun'](x) >= 0 for con in self.constraints]))
        return tmax and tmin and cons



import numpy as np
from scipy.stats import logistic as logistic_dist

from .configuration import Configuration
from .dist import get_dist
from .transform import warp, noise_meta_transform, noise_sens_transform, \
    link_function
from .util import _check_param, TAB, type2roc


class Simulation:
    def __init__(self, nsubjects, nsamples, params, cfg, stimulus_ids, stimuli, choices, dv_sens_prenoise, dv_sens,
                 dv_meta_prenoise=None, dv_meta=None, confidence_prenoise=None, confidence=None, noise_meta=None,
                 likelihood_dist=None):
        self.nsubjects = nsubjects
        self.nsamples = nsamples
        self.params = params
        self.params_sens = {k: v for k, v in self.params.items() if k.endswith('_sens')}
        self.params_meta = {k: v for k, v in self.params.items() if k.endswith('_meta')}
        self.cfg = cfg
        self.stimulus_ids = stimulus_ids
        self.stimuli = stimuli
        self.choices = choices
        self.correct = stimulus_ids == choices
        self.dv_sens_prenoise = dv_sens_prenoise
        self.dv_sens = dv_sens
        self.dv_meta_prenoise = dv_meta_prenoise
        self.dv_meta = dv_meta
        self.confidence_prenoise = confidence_prenoise
        self.confidence = confidence
        self.noise_meta = noise_meta
        self.likelihood_dist = likelihood_dist

    def squeeze(self):
        for var in ('stimulus_ids', 'stimuli', 'choices', 'correct', 'dv_sens', 'dv_meta', 'confidence'):
            if getattr(self, var) is not None:
                setattr(self, var, getattr(self, var).squeeze())
        return self


def generate_stimuli(nsubjects, nsamples, stepsize=0.02):
    levels = np.hstack((-np.arange(stepsize, 1.01, stepsize)[::-1], np.arange(stepsize, 1.01, stepsize)))
    stimuli = np.array([np.random.permutation(np.tile(levels, int(nsamples / len(levels)))) for _ in range(nsubjects)])
    return stimuli


def simu_type1_responses(stimuli, params, cfg):
    if cfg.enable_warping_sens:
        stimuli_final = warp(stimuli, params['warping_sens'], cfg.function_warping_sens)
    else:
        stimuli_final = stimuli
    if cfg.enable_noise_sens:
        noise_sens = noise_sens_transform(
            stimuli=stimuli, function_noise_transform_sens=cfg.function_noise_transform_sens, **params)
    else:
        noise_sens = cfg.noise_sens_default

    thresh_sens = _check_param(params['thresh_sens'] if cfg.enable_thresh_sens else 0)
    bias_sens = _check_param(params['bias_sens'] if cfg.enable_bias_sens else 0)

    dv_sens_prenoise = np.full(stimuli_final.shape, np.nan)
    dv_sens_prenoise[stimuli_final < 0] = (np.abs(stimuli_final[stimuli_final < 0]) > thresh_sens[0]) * \
        stimuli_final[stimuli_final < 0] + bias_sens[0]
    dv_sens_prenoise[stimuli_final >= 0] = (np.abs(stimuli_final[stimuli_final >= 0]) > thresh_sens[1]) * \
        stimuli_final[stimuli_final >= 0] + bias_sens[1]

    if cfg.detection_model:
        nchannels = cfg.detection_model_nchannels  # number of sensory channels
        p_active = np.tanh(np.abs(dv_sens_prenoise) / noise_sens)  # probability that a channel is active
        active = np.repeat(p_active[:, :, None], nchannels, axis=2) > np.random.rand(*dv_sens_prenoise.shape,
                                                                                     nchannels)
        nactive = np.sum(active, axis=2)  # = k
        correct = np.zeros(dv_sens_prenoise.shape)
        correct[(nactive > 0) & (np.sign(dv_sens_prenoise) == np.sign(stimuli_final))] = 1
        correct[(nactive > 0) & (np.sign(dv_sens_prenoise) != np.sign(stimuli_final))] = 0
        correct[nactive == 0] = np.random.randint(0, 2, np.sum(nactive == 0))
        stimulus_ids = (np.sign(stimuli) > 0).astype(int)
        choices = (correct == stimulus_ids).astype(int)
        dv_sens_prenoise = p_active * nchannels
        dv_sens = nactive
        # if the decision value should be in original space, use this:
        # dv_sens = np.sign(dv_sens_prenoise) * noise_transform_sens * \
        #     np.arctanh(np.minimum(nchannels-1e-5, nactive) / nchannels)
    else:
        dv_sens = dv_sens_prenoise + logistic_dist(scale=noise_sens * np.sqrt(3) / np.pi).rvs(size=stimuli.shape)
        choices = (dv_sens >= 0).astype(int)

    return choices, stimuli_final, dv_sens_prenoise, dv_sens


def simu_data(nsubjects, nsamples, params, cfg=None, stimuli_ext=None, verbose=True, stimuli_stepsize=0.02,
              squeeze=False, force_settings=True, **kwargs):
    params = params.copy()  # this variable can be modifed, thus better to make a copy
    if cfg is None:
        # Set configuration attributes that match keyword arguments
        cfg_kwargs = {k: v for k, v in kwargs.items() if k in Configuration.__dict__}
        cfg = Configuration(force_settings=force_settings, **cfg_kwargs)
        for setting in cfg.__dict__:
            if setting.startswith('enable_'):
                if setting.split('enable_')[1] not in params:
                    setattr(cfg, setting, 0)
    cfg.setup()

    if cfg.meta_noise_dist is None:
        cfg.meta_noise_dist = dict(noisy_report='beta', noisy_readout='gamma')[cfg.meta_noise_type]

    if cfg.meta_noise_dist == 'truncated_norm_transform':
        if cfg.meta_noise_type == 'noisy_report':
            lookup_table = np.load('lookup_truncated_norm_noisy_report.npz')
        elif cfg.meta_noise_type == 'noisy_readout':
            lookup_table = np.load('lookup_truncated_norm_noisy_readout.npz')
    else:
        lookup_table = None

    # Make sure no unwanted parameters have been passed
    for p in ('warping', 'thresh', 'bias', 'noise_transform'):
        if not getattr(cfg, f'enable_{p}_sens'):
            params.pop(f'{p}_sens', None)
    for p in ('noise_transform', 'evidence_bias_mult', 'evidence_bias_add', 'evidence_bias_mult_postnoise',
              'confidence_bias_mult', 'confidence_bias_add'):
        if not getattr(cfg, f'enable_{p}_meta'):
            params.pop(f'{p}_meta', None)
    if not cfg.enable_noise_sens:
        params['noise_sens'] = cfg.noise_sens_default
    if not cfg.enable_noise_meta:
        params['noise_meta'] = cfg.noise_meta_default

    if cfg.enable_criteria_meta:
        ncrit_meta = int(cfg.meta_link_function.split('_')[0])
        if 'criteria_meta' not in params:
            if cfg.enable_criteria_meta == 2:
                params['criteria_meta'] = [[params[f'criterion{i}_meta'][j] for i in range(ncrit_meta)] for j in
                                           range(2)]
            else:
                params['criteria_meta'] = [params[f'criterion{i}_meta'] for i in range(ncrit_meta)]
            for i in range(ncrit_meta):
                params.pop(f'criterion{i}_meta', None)
    else:
        for p_ in [p for p in params if p.startswith('criterion')]:
            params.pop(p_, None)
    if cfg.enable_levels_meta:
        ncrit_meta = int(cfg.meta_link_function.split('_')[0])
        if 'levels_meta' not in params:
            if cfg.enable_levels_meta == 2:
                params['levels_meta'] = [[params[f'level{i}_meta'][j] for i in range(ncrit_meta)] for j in range(2)]
            else:
                params['levels_meta'] = [params[f'level{i}_meta'] for i in range(ncrit_meta)]
            for i in range(ncrit_meta):
                params.pop(f'level{i}_meta', None)
    else:
        for p_ in [p for p in params if p.startswith('level')]:
            params.pop(p_, None)

    if stimuli_ext is None:
        stimuli = generate_stimuli(nsubjects, nsamples, stepsize=stimuli_stepsize)
    else:
        stimuli = stimuli_ext / np.max(np.abs(stimuli_ext))
        if stimuli_ext.shape != (nsubjects, nsamples):
            stimuli = np.tile(stimuli, (nsubjects, 1))
    stimulus_ids = (np.sign(stimuli) > 0).astype(int)
    choices, stimuli_final, dv_sens_prenoise, dv_sens = simu_type1_responses(stimuli, params, cfg)

    if not cfg.skip_meta:
        if cfg.enable_evidence_bias_mult_meta == 1:
            dv_meta_prenoise = params['evidence_bias_mult_meta'] * np.abs(dv_sens)
        elif cfg.enable_evidence_bias_mult_meta == 2:
            dv_meta_prenoise = np.full(dv_sens.shape, np.nan)
            neg, pos = dv_sens < 0, dv_sens >= 0
            dv_meta_prenoise[neg] = params['evidence_bias_mult_meta'][0] * np.abs(dv_sens[neg])
            dv_meta_prenoise[pos] = params['evidence_bias_mult_meta'][1] * np.abs(dv_sens[pos])
        else:
            dv_meta_prenoise = np.abs(dv_sens)

        if cfg.enable_evidence_bias_add_meta == 1:
            dv_meta_prenoise = np.maximum(0, dv_meta_prenoise + params['evidence_bias_add_meta'])
        elif cfg.enable_evidence_bias_add_meta == 2:
            dv_meta_prenoise = np.full(dv_sens.shape, np.nan)
            neg, pos = dv_sens < 0, dv_sens >= 0
            dv_meta_prenoise[neg] = np.maximum(0, dv_meta_prenoise[neg] + params['evidence_bias_add_meta'][0])
            dv_meta_prenoise[pos] = np.maximum(0, dv_meta_prenoise[pos] + params['evidence_bias_add_meta'][1])

        if cfg.meta_noise_type == 'noisy_readout':

            if cfg.enable_noise_meta:
                noise_meta = noise_meta_transform(
                    dv_meta_prenoise, dv_sens=dv_sens,
                    function_noise_transform_meta=cfg.function_noise_transform_meta, **params
                )
            else:
                noise_meta = cfg.noise_meta_default
            dist = get_dist(cfg.meta_noise_dist, mode=dv_meta_prenoise, scale=noise_meta,
                            meta_noise_type=cfg.meta_noise_type, lookup_table=lookup_table)  # noqa

            dv_meta = np.maximum(0, dist.rvs((nsubjects, nsamples)))
        else:
            dv_meta = dv_meta_prenoise

        confidence_prenoise = link_function(
            dv_meta=dv_meta, link_fun=cfg.meta_link_function, dv_sens=dv_sens, stimuli=stimuli_final,
            function_noise_transform_sens=cfg.function_noise_transform_sens,
            **params
        )

        if cfg.meta_noise_type == 'noisy_report':
            if cfg.enable_noise_meta:
                noise_meta = noise_meta_transform(
                    confidence_prenoise, dv_sens=dv_sens,
                    function_noise_transform_meta=cfg.function_noise_transform_meta, **params
                )
                noise_meta = np.maximum(cfg.noise_meta_min, noise_meta)
            else:
                noise_meta = cfg.noise_meta_default

            if cfg.meta_noise_dist == 'beta':
                if np.any(noise_meta > 0.5):
                    raise ValueError(f'max(noise_meta) = {np.max(noise_meta):.2f}, but maximum allowed value for '
                                     f'noise_meta is 0.5 for metacognitive type {cfg.meta_noise_type} and noise model '
                                     f'{cfg.meta_noise_dist}')

            dist = get_dist(cfg.meta_noise_dist, mode=confidence_prenoise, scale=noise_meta,
                            meta_noise_type=cfg.meta_noise_type, lookup_table=lookup_table)
            confidence = np.maximum(0, np.minimum(1, dist.rvs((nsubjects, nsamples))))
        else:
            confidence = confidence_prenoise

    if squeeze:
        stimulus_ids = stimulus_ids.squeeze()
        stimuli = stimuli.squeeze()
        choices = choices.squeeze()
        dv_sens_prenoise = dv_sens_prenoise.squeeze()
        dv_sens = dv_sens.squeeze()
        if not cfg.skip_meta:
            dv_meta_prenoise = dv_meta_prenoise.squeeze()  # noqa
            dv_meta = dv_meta.squeeze()  # noqa
            confidence_prenoise = confidence_prenoise.squeeze()  # noqa
            confidence = confidence.squeeze()  # noqa

    simargs = dict(
        nsubjects=nsubjects, nsamples=nsamples, params=params, cfg=cfg,
        stimulus_ids=stimulus_ids, stimuli=stimuli, choices=choices,
        dv_sens_prenoise=dv_sens_prenoise, dv_sens=dv_sens,
    )
    if not cfg.skip_meta:
        simargs.update(
            dv_meta_prenoise=dv_meta_prenoise, dv_meta=dv_meta, confidence_prenoise=confidence_prenoise,  # noqa
            confidence=confidence, noise_meta=noise_meta  # noqa
        )
    simulation = Simulation(**simargs)
    if verbose:
        print('----------------------------------')
        print('Basic stats of the simulated data:')
        correct = (stimulus_ids == choices).astype(int)  # noqa
        print(f'{TAB}Performance: {100 * np.mean(correct):.1f}% correct')
        choice_bias = 100*choices.mean()
        print(f"{TAB}Choice bias: {('-', '+')[int(choice_bias > 50)]}{np.abs(choice_bias - 50):.1f}%")
        if not cfg.skip_meta:
            print(f'{TAB}Confidence: {confidence.mean():.2f}')
            print(f'{TAB}AUROC2: {type2roc(correct, confidence):.2f}')
        print('----------------------------------')

    return simulation


if __name__ == '__main__':
    params_simulation = dict(
        noise_sens=0.2,
        thresh_sens=0.2,
        bias_sens=0.2,
        noise_meta=0.2,
        evidence_bias_add_meta=0
    )
    options = dict(meta_noise_type='noisy_report', enable_thresh_sens=1, enable_bias_sens=1,
                   enable_evidence_bias_add_meta=1)
    m = simu_data(1, 1000, params_simulation, **options)
    # model = reconstruct_observer(m.stimuli[0], m.choices[0], m.confidence[0], **{**options, **params_simulation})
    # plot_sensory_meta(model)



from .model import ReMeta, load_dataset
from .configuration import Configuration
from .gendata import simu_data
from .plot import plot_sensory_meta, plot_psychometric, plot_psychometric_sim, plot_link_function, \
    plot_confidence_dist, plot_confidence, plot_confidence_sim
